# pisek cms - Tool for importing tasks from Pisek into CMS.
#
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from datetime import timedelta
from typing import Optional
from cms.db.task import Task
from sqlalchemy.orm import Session
from sqlalchemy.orm.exc import NoResultFound

from pisek.cms.dataset import create_dataset
from pisek.env.env import Env
from pisek.config.task_config import TaskConfig
from pisek.utils.paths import InputPath


def create_task(
    session: Session,
    env: Env,
    testcases: list[InputPath],
    description: str,
    time_limit: Optional[float],
) -> Task:
    config = env.config

    task = Task(name=config.cms.name, title=config.cms.title)
    set_task_settings(task, config)

    dataset = create_dataset(session, env, task, testcases, description, time_limit)

    task.active_dataset = dataset

    session.add(task)
    return task


def set_task_settings(task: Task, config: TaskConfig):
    task.title = config.cms.title
    task.submission_format = config.cms.submission_format
    task.max_submission_number = config.cms.max_submissions
    task.min_submission_interval = (
        timedelta(seconds=config.cms.min_submission_interval)
        if config.cms.min_submission_interval > 0
        else None
    )
    task.score_precision = config.task.score_precision
    task.score_mode = config.cms.score_mode
    task.feedback_level = config.cms.feedback_level


def get_task(session: Session, config: TaskConfig):
    try:
        return session.query(Task).filter(Task.name == config.cms.name).one()
    except NoResultFound as e:
        raise RuntimeError("This task has not been imported into CMS yet") from e



# pisek cms - Tool for importing tasks from Pisek into CMS.
#
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# The CMS module imports and calls internal CMS libraries,
# as well as dependencies of CMS (sqlalchemy).
# Since these dependencies aren't declared anywhere, importing this module
# will cause an exception on most installations.
# It is therefore important not to import this module unless a cms command has actually been called.

# The implementation uses CMS's sqlalchemy types to make changes to the database.
# It also uses some other APIs, like session creation.
# If CMS changes its database schema or some of its internal APIs,
# this module will likely break.

# This module obviously also uses internal methods of pisek,
# mostly config parsing and some general utilities.
# Large refactors in pisek will also require changes to this module.

# Some functionality is duplicated, most notably enumerating all inputs
# and finding the file extensions of solutions.
# All of this functionality is implemented inside private methods of pipeline jobs,
# making it inaccessible to part of Pisek that don't use the pipeline.

# The pipeline is called when creating a dataset, in order to generate inputs and outputs
# and to compile the judge.
# It is called as if the user requested the primary solution to be tested.

# One small thing to note is that which automatic submission corresponds
# to which solution isn't explicitly stored anywhere.
# Instead, the check and testing-log commands simply search the database for
# a matching (task, file hash, language) triple.

from argparse import Namespace
from cms.db.session import Session
from cms.db.task import Dataset, Task
from sqlalchemy.orm import Session as SessionType
from sqlalchemy.exc import IntegrityError

from pisek.cms.dataset import (
    create_dataset,
    get_dataset_by_description,
    get_only_dataset,
)
from pisek.cms.result import create_testing_log, check_results
from pisek.cms.submission import get_participation, submit_all
from pisek.cms.task import create_task, get_task, set_task_settings
from pisek.env.env import Env, TestingTarget
from pisek.jobs.cache import Cache
from pisek.jobs.task_pipeline import TaskPipeline
from pisek.utils.paths import InputPath
from pisek.utils.pipeline_tools import with_env
from pisek.utils.util import clean_non_relevant_files


def generate_testcases(env: Env) -> list[InputPath]:
    env = env.fork()
    env.solutions = [env.config.primary_solution]
    env.target = TestingTarget.solutions

    pipeline = TaskPipeline(env)

    if pipeline.run_jobs(Cache(), env) != 0:
        raise RuntimeError("Failed to test primary solution, cannot upload to CMS")
    clean_non_relevant_files(pipeline.all_accessed_files)

    return pipeline.input_dataset()


def check_config(env: Env):
    if env.config.cms.name is None:
        raise RuntimeError(
            "The name key in the [cms] section must be set to run CMS commands."
        )


@with_env
def create(env: Env, args: Namespace) -> None:
    check_config(env)
    testcases = generate_testcases(env)
    session = Session()

    description = args.description
    time_limit = args.time_limit

    task = create_task(session, env, testcases, description, time_limit)
    dataset = task.active_dataset

    try:
        session.commit()
    except IntegrityError as e:
        raise RuntimeError(
            "Failed to commit transaction, does the task already exist?"
        ) from e

    print(
        f'Created task {task.name} (id {task.id}) with dataset "{dataset.description}" (id {dataset.id})'
    )


@with_env
def update(env: Env, args: Namespace) -> None:
    check_config(env)
    session = Session()

    task = get_task(session, env.config)
    set_task_settings(task, env.config)

    session.commit()

    print(f"Updated task {task.name} (id {task.id})")


@with_env
def add(env: Env, args: Namespace) -> None:
    check_config(env)
    testcases = generate_testcases(env)
    session = Session()

    description = args.description
    time_limit = args.time_limit
    autojudge = not args.no_autojudge

    task = get_task(session, env.config)
    dataset = create_dataset(
        session, env, task, testcases, description, time_limit, autojudge
    )

    try:
        session.commit()
    except IntegrityError as e:
        raise RuntimeError(
            "Failed to commit transaction, does a dataset with this description exist already?"
        ) from e

    print(f'Added dataset "{dataset.description}" (id {dataset.id})')


@with_env
def submit(env: Env, args: Namespace) -> None:
    check_config(env)
    session = Session()

    username = args.username

    task = get_task(session, env.config)
    participation = get_participation(session, task, username)
    submissions = submit_all(session, env, task, participation)

    session.commit()

    for solution, submission in submissions:
        print(f"Submitted {solution} with id {submission.id}")


def get_dataset_from_args(session: SessionType, task: Task, args: Namespace) -> Dataset:
    if args.active_dataset:
        return task.active_dataset
    elif args.dataset is not None:
        return get_dataset_by_description(session, task, args.dataset)
    else:
        return get_only_dataset(session, task)


@with_env
def testing_log(env: Env, args: Namespace) -> None:
    check_config(env)
    session = Session()

    task = get_task(session, env.config)
    dataset = get_dataset_from_args(session, task, args)
    create_testing_log(session, env, dataset)


@with_env
def check(env: Env, args: Namespace) -> None:
    check_config(env)
    session = Session()

    task = get_task(session, env.config)
    dataset = get_dataset_from_args(session, task, args)
    check_results(session, env, dataset)



# pisek cms - Tool for importing tasks from Pisek into CMS.
#
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional
from cms.db.contest import Contest
from cms.db.task import Task
from cms.db.user import Participation, User
from cms.db.submission import Submission, File
from cms.db.filecacher import FileCacher
from cms.grading.language import Language
from cms.grading.languagemanager import get_language
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm import Session
from os import path
from datetime import datetime, timezone

from pisek.env.env import Env
from pisek.config.task_config import SolutionSection
from pisek.utils.colors import ColorSettings
from pisek.utils.paths import TaskPath
from pisek.utils.text import eprint


def get_participation(session: Session, task: Task, username: str) -> Participation:
    if task.contest_id is None:
        raise RuntimeError("The task is not part of any contest")

    try:
        return (
            session.query(Participation)
            .join(User)
            .filter(Participation.contest_id == task.contest_id)
            .filter(User.username == username)
            .one()
        )
    except NoResultFound as e:
        raise RuntimeError(
            f'There is no user named "{username}" in the given contest'
        ) from e


def submit_all(
    session: Session, env: Env, task: Task, participation: Participation
) -> list[tuple[str, Submission]]:
    config = env.config

    files = FileCacher()
    submissions = []

    for name, solution in config.solutions.items():
        submission = submit(session, files, env, solution, task, participation)
        if submission is not None:
            submissions.append((name, submission))

    return submissions


def submit(
    session: Session,
    files: FileCacher,
    env: Env,
    solution: SolutionSection,
    task: Task,
    participation: Participation,
) -> Submission:
    resolved = resolve_solution(task.contest, env, solution)
    if resolved is None:
        return None
    file_path, language = resolved

    if len(task.submission_format) != 1:
        raise RuntimeError(
            "Cannot submit solutions to tasks that require multiple files"
        )

    digest = files.put_file_from_path(file_path.path, f"Solution to task {task.id}")
    submission = get_submission_of_digest(session, digest, language, task)

    if submission is not None:
        return submission

    submission = Submission(
        timestamp=datetime.now(timezone.utc).replace(tzinfo=None),
        opaque_id=Submission.generate_opaque_id(session, participation.id),
        language=language.name,
        participation=participation,
        task=task,
        comment=solution.name,
    )
    session.add(submission)

    filename = task.submission_format[0]
    session.add(File(filename=filename, digest=digest, submission=submission))

    return submission


def get_submission(
    session: Session,
    files: FileCacher,
    env: Env,
    solution: SolutionSection,
    task: Task,
) -> Optional[Submission]:
    if task.contest is None:
        raise RuntimeError("The task is not part of any contest")

    resolved = resolve_solution(task.contest, env, solution)
    if resolved is None:
        return None
    file_path, language = resolved

    digest = files.put_file_from_path(file_path.path, f"Solution to task {task.name}")

    return get_submission_of_digest(session, digest, language, task)


def get_submission_of_digest(
    session: Session,
    digest: str,
    language: Language,
    task: Task,
) -> Optional[Submission]:
    return (
        session.query(Submission)
        .join(File)
        .filter(Submission.task == task)
        .filter(Submission.language == language.name)
        .filter(File.digest == digest)
        .order_by(Submission.timestamp.desc())
        .first()
    )


def resolve_solution(
    contest: Contest, env: Env, solution: SolutionSection
) -> Optional[tuple[TaskPath, Language]]:
    sources = solution.run.build.sources
    if len(sources) != 1:
        eprint(
            ColorSettings.colored(
                "For CMS import there must be exactly one source:\n  "
                + " ".join(map(lambda p: p.col(env), sources)),
                "yellow",
            )
        )
        return None

    source: TaskPath = sources[0]

    for language_name in contest.languages:
        language: Language = get_language(language_name)

        for ext in language.source_extensions:
            if source.name.endswith(ext):
                return source, language

            file_path = source.replace_suffix(ext)

            if path.isfile(file_path.path):
                return file_path, language

    eprint(
        ColorSettings.colored(
            f"Skipping {source}, as it isn't available in any enabled language",
            "yellow",
        )
    )
    return None



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2025        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


class UserError(Exception):
    pass


class TaskConfigError(UserError):
    pass


class TaskConfigParsingError(TaskConfigError):
    def __init__(self, path: str, msg: str) -> None:
        super().__init__(f"Unable to parse '{path}': {msg}")


class MissingFile(UserError):
    pass


class TestingFailed(UserError):
    pass


class InvalidArgument(UserError):
    pass


class InvalidOperation(UserError):
    pass


class NotSupported(UserError):
    pass



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from dataclasses import dataclass
import os
from typing import TYPE_CHECKING
from pisek.config.config_types import DataFormat

if TYPE_CHECKING:
    from pisek.env.env import Env

BUILD_DIR = "build/"
TESTS_DIR = "tests/"
INTERNALS_DIR = ".pisek/"

GENERATED_SUBDIR = "_generated/"
INPUTS_SUBDIR = "_inputs/"
FUZZING_OUTPUTS_SUBDIR = "_fuzzing/"

INPUTS_LIST = "_inputs_list"


@dataclass(frozen=True)
class TaskPath:
    """Class representing a path to task file."""

    path: str

    def __init__(self, *path: str):
        joined_path = os.path.normpath(os.path.join(*path))
        object.__setattr__(self, "path", joined_path)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={self.path})"

    def __init_subclass__(cls):
        return super().__init_subclass__()

    @property
    def name(self) -> str:
        return os.path.basename(self.path)

    @property
    def abspath(self) -> str:
        return os.path.abspath(self.path)

    def __format__(self, __format_spec: str) -> str:
        match __format_spec:
            case "":
                return self.path
            case "p":
                return self.path
            case "n":
                return self.name
            case _:
                raise ValueError(
                    f"Invalid format specifier '{__format_spec}' for object of type '{self.__class__.__name__}'"
                )

    def __eq__(self, other_path) -> bool:
        if isinstance(other_path, TaskPath):
            return self.path == other_path.path
        else:
            return False

    def __lt__(self, other_path: "TaskPath") -> bool:
        return self.path < other_path.path

    def col(self, env: "Env") -> str:
        return env.colored(
            self.path + ("/" if os.path.isdir(self.path) else ""), "magenta"
        )

    def replace_suffix(self, new_suffix: str) -> "TaskPath":
        path = os.path.splitext(self.path)[0] + new_suffix
        return TaskPath(path)

    def join(self, *path: str) -> "TaskPath":
        return TaskPath(os.path.join(self.path, *path))

    def exists(self) -> bool:
        return os.path.exists(self.path)

    @staticmethod
    def from_abspath(*path: str) -> "TaskPath":
        return TaskPath(os.path.relpath(os.path.join(*path), "."))

    @staticmethod
    def static_path(env: "Env", *path: str) -> "TaskPath":
        return env.config.tests.static_subdir.join(*path)

    @staticmethod
    def executable_path(*path: str) -> "TaskPath":
        return TaskPath(BUILD_DIR, *path)

    @staticmethod
    def data_path(*path: str) -> "TaskPath":
        return TaskPath(TESTS_DIR, *path)

    @staticmethod
    def generated_path(*path: str) -> "TaskPath":
        return TaskPath.data_path(GENERATED_SUBDIR, *path)


# ----- interfaces -----


class IJudgeablePath(TaskPath, ABC):
    @abstractmethod
    def to_checker_log(self, judge: str) -> "LogPath":
        pass


class ISanitizedPath(TaskPath, ABC):
    @abstractmethod
    def to_raw(self, format: DataFormat) -> "IRawPath":
        pass


class IInputPath(ISanitizedPath):
    @abstractmethod
    def to_second(self) -> "IInputPath":
        pass

    @abstractmethod
    def to_output(self) -> "IOutputPath":
        pass

    @abstractmethod
    def to_log(self, program: str) -> "LogPath":
        pass


class IOutputPath(IJudgeablePath, ISanitizedPath):
    @abstractmethod
    def to_reference_output(self) -> "IOutputPath":
        pass

    @abstractmethod
    def to_fuzzing(self, note: str, seed: int) -> "IOutputPath":
        pass


class IRawPath(TaskPath, ABC):
    @abstractmethod
    def to_sanitized_input(self) -> IInputPath:
        pass

    @abstractmethod
    def to_sanitized_output(self) -> IOutputPath:
        pass

    @abstractmethod
    def to_sanitization_log(self) -> "LogPath":
        pass


# ----- paths inside task -----


class JudgeablePath(IJudgeablePath):
    def to_checker_log(self, judge: str) -> "LogPath":
        return LogPath(self.replace_suffix(f".{os.path.basename(judge)}.log").path)


class SanitizedPath(ISanitizedPath):
    def to_raw(self, format: DataFormat) -> "RawPath":
        if format == DataFormat.binary:
            return RawPath(self.path)
        return RawPath(self.path + ".raw")


class InputPath(SanitizedPath, IInputPath):
    @staticmethod
    def new(*path: str, solution: str | None = None) -> "InputPath":
        if solution is None:
            return InputPath(TESTS_DIR, INPUTS_SUBDIR, *path)
        else:
            return InputPath(TESTS_DIR, solution, *path)

    def to_second(self) -> IInputPath:
        return InputPath(self.replace_suffix(".in2").path)

    def to_output(self) -> IOutputPath:
        return OutputPath(self.replace_suffix(f".out").path)

    def to_log(self, program: str) -> "LogPath":
        return LogPath(self.replace_suffix(f".{os.path.basename(program)}.log").path)


class OutputPath(JudgeablePath, SanitizedPath, IOutputPath):
    @staticmethod
    def static(*path) -> "OutputPath":
        return OutputPath(TESTS_DIR, INPUTS_SUBDIR, *path)

    def to_reference_output(self) -> IOutputPath:
        return OutputPath(self.replace_suffix(f".ok").path)

    def to_fuzzing(self, note: str, seed: int) -> IOutputPath:
        return OutputPath(
            TESTS_DIR,
            FUZZING_OUTPUTS_SUBDIR,
            self.replace_suffix(f".{note}.{seed:x}.out").name,
        )


class LogPath(JudgeablePath):
    @staticmethod
    def generator_log(generator: str) -> "LogPath":
        return LogPath(TESTS_DIR, INPUTS_SUBDIR, f"{os.path.basename(generator)}.log")


class RawPath(IRawPath):
    def to_sanitized_input(self) -> IInputPath:
        return InputPath(self.path.removesuffix(".raw"))

    def to_sanitized_output(self) -> IOutputPath:
        return OutputPath(self.path.removesuffix(".raw"))

    def to_sanitization_log(self) -> LogPath:
        return LogPath(self.replace_suffix(".sanitizer.log").path)


# ----- opendata paths = paths outside task -----


class OpendatadPath(TaskPath):
    def __init__(self, tmp_dir: str, *path: str):
        self._tmp_dir = tmp_dir
        super().__init__(*path)


class OpendataJudgeablePath(OpendatadPath, IJudgeablePath):
    def to_checker_log(self, judge: str) -> "LogPath":
        return LogPath(
            self._tmp_dir, self.replace_suffix(f".{os.path.basename(judge)}.log").name
        )


class OpendataSanitizedPath(OpendatadPath, ISanitizedPath):
    def to_raw(self, format: DataFormat) -> IRawPath:
        if format == DataFormat.binary:
            return OpendataRawPath(self._tmp_dir, self.path)
        return RawPath(self._tmp_dir, self.name + ".raw")


class OpendataInputPath(OpendataSanitizedPath, IInputPath):
    def to_second(self) -> IInputPath:
        return InputPath(self._tmp_dir, self.replace_suffix(".in2").name)

    def to_output(self) -> IOutputPath:
        return OutputPath(self._tmp_dir, self.replace_suffix(f".out").name)

    def to_log(self, program: str) -> "LogPath":
        return LogPath(
            self._tmp_dir, self.replace_suffix(f".{os.path.basename(program)}.log").name
        )


class OpendataOutputPath(OpendataSanitizedPath, OpendataJudgeablePath, IOutputPath):
    def to_reference_output(self) -> IOutputPath:
        return OutputPath(self._tmp_dir, self.replace_suffix(f".ok").name)

    def to_fuzzing(self, note: str, seed: int) -> IOutputPath:
        return OutputPath(
            self._tmp_dir,
            self.replace_suffix(f".{note}.{seed:x}.out").name,
        )


class OpendataRawPath(OpendataSanitizedPath, IRawPath):
    def to_sanitized_input(self) -> IInputPath:
        return InputPath(self._tmp_dir, self.name.removesuffix(".raw"))

    def to_sanitized_output(self) -> IOutputPath:
        return OutputPath(self._tmp_dir, self.name.removesuffix(".raw"))

    def to_sanitization_log(self) -> LogPath:
        return LogPath(self._tmp_dir, self.replace_suffix(".sanitizer.log").name)



import os
import unittest

from util import TestFixtureVariant


class TestMax(TestFixtureVariant):
    @property
    def fixture_path(self) -> str:
        return "fixtures/max/"

    def set_env(self):
        os.environ["DEBUG"] = "false"


if __name__ == "__main__":
    unittest.main(verbosity=2)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pisek.utils.paths import IInputPath, IOutputPath
from pisek.config.task_config import TestSection
from pisek.jobs.status import StatusJobManager
from pisek.task_jobs.task_job import TaskHelper
from pisek.task_jobs.data.testcase_info import TestcaseInfo


TOOLS_MAN_CODE = "tools"
GENERATOR_MAN_CODE = "generator"
INPUTS_MAN_CODE = "inputs"
BUILD_MAN_CODE = "build"
SOLUTION_MAN_CODE = "solution_"
DATA_MAN_CODE = "data"
FUZZ_MAN_CODE = "fuzz"


class TaskJobManager(StatusJobManager, TaskHelper):
    """JobManager class that implements useful methods"""

    def _get_samples(self) -> list[tuple[IInputPath, IOutputPath]]:
        """Returns the list [(sample1.in, sample1.out), …]."""
        return [
            (
                inp.input_path(),
                inp.reference_output(self._env),
            )
            for inp in self._test_testcases(self._env.config.test_sections[0])
        ]

    def _test_testcases(self, test: TestSection) -> list[TestcaseInfo]:
        """Get all inputs of given test."""
        return self.prerequisites_results[INPUTS_MAN_CODE]["testcase_info"][test.num]

    def _all_testcases(self) -> dict[int, list[TestcaseInfo]]:
        """Get all inputs grouped by test."""
        return self.prerequisites_results[INPUTS_MAN_CODE]["testcase_info"]



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
from typing import NoReturn

from pisek.utils.colors import ColorSettings, remove_colors


def tab(text: str, tab_str: str = "  ") -> str:
    return tab_str + text.replace("\n", f"\n{tab_str}")


def pad(text: str, length: int, pad_char: str = " ") -> str:
    return text + (length - len(remove_colors(text))) * pad_char


def pad_left(text: str, length: int, pad_char: str = " ") -> str:
    return pad(text[::-1], length, pad_char)[::-1]


def eprint(msg, *args, **kwargs) -> None:
    """Prints to sys.stderr."""
    print(msg, *args, file=sys.stderr, **kwargs)


def fatal_user_error(msg, *args, **kwargs) -> NoReturn:
    eprint(ColorSettings.colored(msg, "red"), *args, **kwargs)
    exit(2)


def warn(msg: str, err: type, strict: bool = False) -> None:
    """Warn if strict is False, otherwise raise error."""
    if strict:
        raise err(msg)
    eprint(ColorSettings.colored(msg, "yellow"))



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from importlib.resources import files
import os

from pisek.jobs.jobs import Job, PipelineItemFailure
from pisek.config.config_types import DataFormat
from pisek.env.env import Env
from pisek.utils.paths import TaskPath, IRawPath, ISanitizedPath
from pisek.task_jobs.task_job import TaskJob
from pisek.task_jobs.task_manager import TaskJobManager
from pisek.task_jobs.program import ProgramsJob
from pisek.task_jobs.run_result import RunResult, RunResultKind


class ToolsManager(TaskJobManager):
    """Manager that prepares all tools necessary for task testing."""

    def __init__(self):
        super().__init__("Preparing tools")

    def _get_jobs(self) -> list[Job]:
        self.makedirs(TaskPath.executable_path("."))
        jobs: list[Job] = [
            PrepareMinibox(self._env),
            PrepareTextPreprocessor(self._env),
        ]
        return jobs


class PrepareMinibox(TaskJob):
    """Compiles minibox."""

    def __init__(self, env: Env, **kwargs) -> None:
        super().__init__(env=env, name="Prepare Minibox", **kwargs)

    def _run(self):
        source = files("pisek").joinpath("tools/minibox.c")
        executable = TaskPath.executable_path("_minibox")
        self._access_file(executable)
        gcc = self._run_subprocess(
            [
                "gcc",
                source,
                "-o",
                executable.path,
                "-std=gnu11",
                "-D_GNU_SOURCE",
                "-O2",
                "-Wall",
                "-Wextra",
                "-Wno-parentheses",
                "-Wno-sign-compare",
                "-Wno-unused-result",
            ]
        )
        if gcc.returncode != 0:
            raise PipelineItemFailure("Minibox compilation failed.")


class PrepareTextPreprocessor(TaskJob):
    """Compiles Text Preprocessor."""

    def __init__(self, env: Env, **kwargs) -> None:
        super().__init__(env=env, name="Prepare text preprocessor", **kwargs)

    def _run(self):
        source = files("pisek").joinpath("tools/text-preproc.c")
        executable = TaskPath.executable_path("_text-preproc")
        self._access_file(executable)
        gcc = self._run_subprocess(
            [
                "gcc",
                source,
                "-o",
                executable.path,
                "-std=gnu11",
                "-O2",
                "-Wall",
                "-Wextra",
                "-Wno-parentheses",
                "-Wno-sign-compare",
            ]
        )
        if gcc.returncode != 0:
            raise PipelineItemFailure("Text preprocessor compilation failed.")


class PrepareJudgeLibChecker(TaskJob):
    """Compiles judge from judgelib."""

    def __init__(self, env: Env, checker_name: str, judge: str, **kwargs) -> None:
        self.judge = judge
        super().__init__(env=env, name=f"Prepare {checker_name}", **kwargs)

    def _run(self):
        source_files = ["util.cc", "io.cc", "token.cc", "random.cc", f"{self.judge}.cc"]
        source_dir = files("pisek").joinpath("tools/judgelib")
        sources = [source_dir.joinpath(file) for file in source_files]

        executable = TaskPath.executable_path("_" + self.judge)
        self._access_file(executable)

        gpp = self._run_subprocess(
            [
                "g++",
                *sources,
                "-I",
                source_dir,
                "-o",
                executable.path,
                "-std=gnu++17",
                "-O2",
                "-Wall",
                "-Wextra",
                "-Wno-parentheses",
                "-Wno-sign-compare",
            ]
        )

        if gpp.returncode != 0:
            raise PipelineItemFailure(f"{self.checker_name} compilation failed.")


class PrepareTokenJudge(PrepareJudgeLibChecker):
    """Compiles judge-token from judgelib."""

    def __init__(self, env: Env, **kwargs) -> None:
        super().__init__(
            env=env, checker_name="token judge", judge="judge-token", **kwargs
        )


class PrepareShuffleJudge(PrepareJudgeLibChecker):
    """Compiles judge-shuffle from judgelib."""

    def __init__(self, env: Env, **kwargs) -> None:
        super().__init__(
            env=env, checker_name="shuffle judge", judge="judge-shuffle", **kwargs
        )


class SanitizationResultKind(Enum):
    ok = auto()
    changed = auto()
    invalid = auto()
    skipped = auto()


@dataclass(frozen=True)
class SanitizationResult:
    kind: SanitizationResultKind
    msg: str | None = None


class TextPreprocAbstract(ProgramsJob):
    """Abstract job that has method for file sanitization."""

    def _run_text_preproc(
        self, input_: IRawPath, output: ISanitizedPath
    ) -> SanitizationResult:
        try:
            os.remove(output.path)
        except FileNotFoundError:
            pass

        result = self._run_tool(
            "_text-preproc",
            stdin=input_,
            stdout=output,
            stderr=input_.to_sanitization_log(),
        )
        msg = self._read_file(result.stderr_file).strip()
        if result.returncode == 42:
            if self._files_equal(input_, output):
                return SanitizationResult(SanitizationResultKind.ok, msg)
            else:
                return SanitizationResult(SanitizationResultKind.changed, msg)
        elif result.returncode == 43:
            return SanitizationResult(SanitizationResultKind.invalid, msg)
        else:
            raise RuntimeError(f"Text preprocessor failed on file {input_:p}.")


class SanitizeAbstract(TaskJob):
    def __init__(self, env: Env, input_: TaskPath, output: TaskPath, **kwargs) -> None:
        self.input = input_
        self.output = output
        super().__init__(env=env, **kwargs)

    def _run(self) -> SanitizationResult:
        result = self.prerequisites_results.get("create_source", None)
        if isinstance(result, RunResult) and result.kind != RunResultKind.OK:
            self._copy_file(self.input, self.output)
            return SanitizationResult(SanitizationResultKind.skipped)

        return self._sanitize()

    @abstractmethod
    def _sanitize(self) -> SanitizationResult:
        pass


class Sanitize(SanitizeAbstract, TextPreprocAbstract):
    """Sanitize text file using Text Preprocessor."""

    def __init__(
        self, env: Env, input_: IRawPath, output: ISanitizedPath, **kwargs
    ) -> None:
        super().__init__(
            env, input_, output, name=f"Sanitize {input_:p} -> {output:p}", **kwargs
        )

    def _sanitize(self):
        return self._run_text_preproc(self.input, self.output)


class IsClean(SanitizeAbstract, TextPreprocAbstract):
    """Check that file is same after sanitizing with Text Preprocessor."""

    def __init__(
        self, env: Env, input_: IRawPath, output: ISanitizedPath, **kwargs
    ) -> None:
        super().__init__(
            env, input_, output, name=f"Check {input_:p} is clean", **kwargs
        )

    def _sanitize(self):
        result = self._run_text_preproc(self.input, self.output)
        if result.kind != SanitizationResultKind.ok:
            raise PipelineItemFailure(
                f"File {self.input:p} not normalized. (Check encoding, missing newline at the end or '\\r'.)"
            )


def _sanitize_get_format(env: Env, is_input: bool) -> DataFormat:
    if is_input:
        return env.config.tests.in_format
    else:
        return env.config.tests.out_format


def sanitize_job(env: Env, path: ISanitizedPath, is_input: bool) -> Job | None:
    return sanitize_job_direct(
        env, path.to_raw(_sanitize_get_format(env, is_input)), path, is_input
    )


def sanitize_job_direct(
    env: Env, path_from: IRawPath, path_to: ISanitizedPath, is_input: bool
) -> Job | None:
    format_ = _sanitize_get_format(env, is_input)

    if format_ == DataFormat.binary:
        return None
    elif format_ == DataFormat.strict_text and is_input:
        return IsClean(env, path_from, path_to)
    else:
        return Sanitize(env, path_from, path_to)



#!/usr/bin/env python3
import sys

test = int(sys.argv[1])
n, k = map(int, input().split(" "))  # We use explicitly split(" ") to be more strict

assert 1 <= n <= 10**5, f"{n=} limits"
assert 1 <= k <= 10**9, f"{k=} limits"

if test == 1:
    assert n == k, "N and K should be equal"
elif test == 2:
    assert k <= 10**5, f"{k=} is too big"

try:
    input()
except EOFError:
    exit(42)  # The input is valid

assert False, "The input doesn't end when it should"



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from contextlib import contextmanager
from contextvars import ContextVar
from pydantic import BaseModel

from typing import Any, Iterator


_init_context_var: ContextVar[dict[str, Any]] = ContextVar(
    "_init_context_var", default={}
)


@contextmanager
def init_context(value: dict[str, Any]) -> Iterator[None]:
    token = _init_context_var.set(value)
    try:
        yield
    finally:
        _init_context_var.reset(token)


class ContextModel(BaseModel):
    """Model that adds context when validating at model initialization."""

    def __init__(self, /, **data: Any) -> None:
        __tracebackhide__ = True
        self.__pydantic_validator__.validate_python(
            data,
            self_instance=self,
            context=_init_context_var.get(),
        )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from pisek.config.task_config import TaskConfig
from pisek.task_jobs.solution.solution_result import (
    Verdict,
    TEST_SPEC,
    verdict_always,
)


def check_verdicts(
    verdicts: list[Verdict], expected: str
) -> tuple[bool, bool, Optional[int]]:
    """
    Returns tuple:
        - whether verdicts are as expected
        - whether the result is definitive (cannot be changed)
        - a verdict that makes the result the way it is (if there is one particular)
    """
    result = True
    definitive = True
    guarantor = None

    oks = list(map(TEST_SPEC[expected][0], verdicts))
    ok = all(oks)
    result &= ok
    definitive &= not ok or TEST_SPEC[expected][0] == verdict_always
    if not ok:
        return result, definitive, oks.index(False)

    oks = list(map(TEST_SPEC[expected][1], verdicts))
    ok = any(oks)
    result &= ok
    definitive &= ok
    if not ok and len(verdicts) == 1:
        guarantor = 0
    elif ok and TEST_SPEC[expected][1] != verdict_always:
        guarantor = oks.index(True)

    return result, definitive, guarantor


def compute_verdict(verdicts: list[Verdict]) -> Verdict:
    return max(verdicts, default=Verdict.ok, key=lambda v: v.value)



import unittest

from util import TestFixtureVariant


class TestGuess(TestFixtureVariant):
    @property
    def fixture_path(self) -> str:
        return "fixtures/guess/"


if __name__ == "__main__":
    unittest.main(verbosity=2)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import tempfile
import time
from typing import Optional

from pisek.env.env import Env
from pisek.jobs.jobs import State
from pisek.utils.paths import IInputPath, IOutputPath
from pisek.config.config_types import ProgramRole
from pisek.config.task_config import RunSection
from pisek.task_jobs.program import RunResult, ProgramsJob
from pisek.task_jobs.solution.solution_result import Verdict, SolutionResult
from pisek.task_jobs.checker.cms_judge import RunCMSJudge


class RunSolution(ProgramsJob):
    """Runs solution on given input."""

    def __init__(
        self,
        env: Env,
        name: str,
        solution: RunSection,
        is_primary: bool,
        **kwargs,
    ) -> None:
        super().__init__(env=env, name=name, **kwargs)
        self._needed_by = 1
        self.solution = solution
        self.is_primary = is_primary

    def require(self):
        self._needed_by += 1

    def unrequire(self):
        self._needed_by -= 1
        if self._needed_by == 0 and self.state == State.in_queue:
            self.cancel()
        assert self._needed_by >= 0

    def _solution_type(self) -> ProgramRole:
        return (
            (ProgramRole.primary_solution)
            if self.is_primary
            else (ProgramRole.secondary_solution)
        )


class RunBatchSolution(RunSolution):
    def __init__(
        self,
        env: Env,
        solution: RunSection,
        is_primary: bool,
        input_: IInputPath,
        output: IOutputPath,
        **kwargs,
    ) -> None:
        super().__init__(
            env=env,
            name=f"Run {solution.name} on input {input_:n}",
            solution=solution,
            is_primary=is_primary,
            **kwargs,
        )
        self.input = input_
        self.output = output.to_raw(env.config.tests.out_format)
        self.log_file = input_.to_log("solution")

    def _run(self) -> RunResult:
        return self._run_program(
            program_role=self._solution_type(),
            program=self.solution,
            stdin=self.input,
            stdout=self.output,
            stderr=self.log_file,
        )


class RunInteractive(RunCMSJudge, RunSolution):
    def __init__(
        self,
        env: Env,
        solution: RunSection,
        is_primary: bool,
        judge: RunSection,
        test: int,
        input_: IInputPath,
        expected_verdict: Optional[Verdict] = None,
        **kwargs,
    ):
        self.sol_log_file = input_.to_log("solution")
        super().__init__(
            env=env,
            name=f"Run {solution.name} on input {input_:n}",
            judge=judge,
            test=test,
            input_=input_,
            expected_verdict=expected_verdict,
            checker_log_file=self.sol_log_file.to_checker_log(judge.name),
            solution=solution,
            is_primary=is_primary,
            **kwargs,
        )

    def _get_solution_run_res(self) -> RunResult:
        with tempfile.TemporaryDirectory() as fifo_dir:
            fifo_from_solution = os.path.join(fifo_dir, "solution-to-manager")
            fifo_to_solution = os.path.join(fifo_dir, "manager-to-solution")

            os.mkfifo(fifo_from_solution)
            os.mkfifo(fifo_to_solution)

            pipes = [
                os.open(fifo_from_solution, os.O_RDWR),
                os.open(fifo_to_solution, os.O_RDWR),
                # Open fifos to prevent blocking on future opens
                fd_from_solution := os.open(fifo_from_solution, os.O_WRONLY),
                fd_to_solution := os.open(fifo_to_solution, os.O_RDONLY),
            ]

            self._load_program(
                ProgramRole.judge,
                self.judge,
                stdin=self.input,
                stdout=self.points_file,
                stderr=self.checker_log_file,
                args=[fifo_from_solution, fifo_to_solution],
            )

            self._load_program(
                self._solution_type(),
                self.solution,
                stdin=fd_to_solution,
                stdout=fd_from_solution,
                stderr=self.sol_log_file,
            )

            def close_pipes(_):
                time.sleep(0.05)
                for pipe in pipes:
                    os.close(pipe)

            self._load_callback(close_pipes)

            judge_res, sol_res = self._run_programs()
            self._judge_run_result = judge_res

            return sol_res

    def _check(self) -> SolutionResult:
        return self._load_solution_result(self._judge_run_result)

    def _checking_message(self) -> str:
        return f"solution {self.solution.name} on input {self.input.col(self._env)}"



#!/usr/bin/env python3
import os
import sys
import subprocess

inp = sys.stdin.read()
program_dir = os.path.dirname(sys.argv[0])
subprocess.run(os.path.join(program_dir, "solve"), input=inp.encode())



#!/usr/bin/env python3
import os
import sys
from typing import NoReturn

SUBTASK_POINTS = [0, 20, 30, 50]


def award(points: float | None, msg: str) -> NoReturn:
    assert "\n" not in msg
    print(msg, file=sys.stderr)
    print(f"POINTS={points}", file=sys.stderr)
    exit(42)


def reject(msg: str) -> NoReturn:
    assert "\n" not in msg
    print(msg, file=sys.stderr)
    exit(43)


def main(subtask: int, seed: str) -> NoReturn:
    # Read the test input
    with open(os.environ["TEST_INPUT"]) as f:
        n, k = map(int, f.readline().split())

    # Read the contestant output
    try:
        numbers = list(map(int, input().split()))
    except ValueError:
        reject("The output should contain integers.")
    except EOFError:
        reject("The output is empty.")

    try:
        input()
        reject("The output should contain only one line.")
    except EOFError:
        pass

    # Be careful to check **ALL** constraints
    if any(map(lambda x: x <= 0, numbers)):
        reject("The output contains negative integers.")

    if sum(numbers) == k:
        award(SUBTASK_POINTS[subtask], "All of the sand used.")
    elif sum(numbers) >= k / 2:
        award(SUBTASK_POINTS[subtask] // 2, "At least half of the sand used.")
    else:
        reject("Not enough sand used.")


if __name__ == "__main__":
    main(subtask=int(sys.argv[1]), seed=sys.argv[2])



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>
# Copyright (c)   2025        Antonín Maloň <git@tonyl.eu>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
import logging
import inspect
import subprocess
import os
import json
import shutil
from typing import Any, IO, Optional, Protocol, TYPE_CHECKING

from pisek.utils.text import tab
from pisek.jobs.jobs import PipelineItemFailure
from pisek.config.config_types import BuildStrategyName

if TYPE_CHECKING:
    from pisek.env.env import Env
    from pisek.config.task_config import BuildSection

logger = logging.getLogger(__name__)

ALL_STRATEGIES: dict[BuildStrategyName, type["BuildStrategy"]] = {}


class FakeChangedCWD:
    def __init__(self, strategy: "BuildStrategy", path: str):
        self._strategy = strategy
        self.new = strategy._path(path)

    def __enter__(self):
        self.old = self._strategy.workdir
        self._strategy.workdir = self.new

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._strategy.workdir = self.old


class RunPopen(Protocol):
    def __call__(
        self, args: list[str], stdout: int, stderr: int, text: bool, cwd: str | None
    ) -> subprocess.Popen: ...


class BuildStrategy(ABC):
    name: BuildStrategyName
    extra_sources: Optional[str] = None
    extra_nonsources: Optional[str] = None

    def __init__(
        self,
        build_section: "BuildSection",
        env: "Env",
        _run_subprocess: RunPopen,
    ) -> None:
        self._build_section = build_section
        self._env = env
        self.stderr_output = ""
        self._run_popen = _run_subprocess

        self.workdir = "."

    def __init_subclass__(cls):
        if not inspect.isabstract(cls):
            ALL_STRATEGIES[cls.name] = cls
        return super().__init_subclass__()

    def _print_stderr(self, msg) -> None:
        self.stderr_output += msg + "\n"

    @classmethod
    @abstractmethod
    def applicable_on_files(cls, build: "BuildSection", sources: list[str]) -> bool:
        pass

    @classmethod
    @abstractmethod
    def applicable_on_directory(cls, build: "BuildSection", directory: str) -> bool:
        pass

    @classmethod
    def applicable(cls, build: "BuildSection", sources: list[str]) -> bool:
        directories = any(os.path.isdir(s) for s in sources)
        if not directories:
            return cls.applicable_on_files(build, sources)
        elif len(sources) == 1:
            return cls.applicable_on_directory(build, sources[0])
        else:
            return False

    def build(self, directory: str, sources: list[str], extras: list[str]) -> str:
        self.inputs = os.listdir(directory)
        self.sources = sources
        self.extras = extras
        self.target = "_" + os.path.basename(self._build_section.program_name)
        with FakeChangedCWD(self, directory):
            return self._build()

    @abstractmethod
    def _build(self) -> str:
        pass

    @classmethod
    def _ends_with(cls, source: str, suffixes: list[str]) -> bool:
        return any(source.endswith(suffix) for suffix in suffixes)

    @classmethod
    def _all_end_with(cls, sources: list[str], suffixes: list[str]) -> bool:
        return all(cls._ends_with(source, suffixes) for source in sources)

    # ---- fake cwd utils ----

    def _path(self, path: str) -> str:
        return os.path.join(self.workdir, path)

    def _exists(self, path: str) -> bool:
        return os.path.exists(self._path(path))

    def _isdir(self, path: str) -> bool:
        return os.path.isdir(self._path(path))

    def _listdir(self, path: str | None = None) -> list[str]:
        return os.listdir(self.workdir if path is None else self._path(path))

    def _stat(self, path: str) -> os.stat_result:
        return os.stat(self._path(path))

    def _open(self, file: str, mode: str = "r", newline: str | None = None) -> IO[Any]:
        return open(self._path(file), mode=mode, newline=newline)

    def _chmod(self, path: str, mode: int) -> None:
        os.chmod(self._path(path), mode)

    def _copy(self, src: str, dst: str, follow_symlinks: bool = True) -> None:
        shutil.copy(self._path(src), self._path(dst), follow_symlinks=follow_symlinks)

    def _symlink(self, src: str, dst: str) -> None:
        os.symlink(src, self._path(dst))

    def _makedirs(self, name: str, exist_ok: bool = False) -> None:
        os.makedirs(self._path(name), exist_ok=exist_ok)

    # ---- build helpers ----

    def _load_shebang(self, program: str) -> str:
        """Load shebang from program."""
        with self._open(program, "r", newline="\n") as f:
            first_line = f.readline()

        if not first_line.startswith("#!"):
            raise PipelineItemFailure(f"Missing shebang in {program}")
        if first_line.endswith("\r\n"):
            raise PipelineItemFailure(f"First line ends with '\\r\\n' in {program}")

        return first_line.strip().lstrip("#!")

    def _check_tool(self, tool: str) -> None:
        """Checks that a tool exists."""
        try:
            # XXX: We should technically use self._run_popen but that doesn't implement timeout properly
            # But as we set it to zero, it doesn't matter
            #
            # Also tool.split() because some tools have more parts (e.g. '/usr/bin/env python3')
            subprocess.run(
                tool.split(),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                timeout=0,
                cwd=self.workdir,
            )
        except subprocess.TimeoutExpired:
            pass
        except FileNotFoundError:
            raise PipelineItemFailure(f"Missing tool: {tool}")

    def _run_subprocess(self, args: list[str], program: str, **kwargs) -> str:
        self._check_tool(args[0])

        logger.debug("Building '" + " ".join(args) + "'")
        comp = self._run_popen(
            args,
            **kwargs,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            cwd=self.workdir,
        )

        assert comp.stderr is not None
        assert comp.stdout is not None

        stderr: str = comp.stderr.read()
        if stderr.strip():
            self._print_stderr(stderr)

        if comp.returncode != 0:
            raise PipelineItemFailure(
                f"Build of {program} failed.\n"
                + tab(self._env.colored("> " + " ".join(args), "magenta"))
                + "\n"
                + tab(self._env.colored(stderr, "yellow"))
            )
        return comp.stdout.read()

    def _get_entrypoint(self, file_extension: str) -> str:
        assert file_extension[0] == "."
        if len(self.sources) == 1:
            return self.sources[0]
        else:
            if self._build_section.entrypoint == "":
                raise PipelineItemFailure(
                    f"For multiple {self.name} files 'entrypoint' must be set (in section [{self._build_section.section_name}])."
                )
            if (
                entrypoint := self._build_section.entrypoint + file_extension
            ) in self.sources:
                return entrypoint
            elif (entrypoint := self._build_section.entrypoint) in self.sources:
                return entrypoint
            else:
                raise PipelineItemFailure(
                    f"Entrypoint '{self._build_section.entrypoint}' not in sources."
                )

    def _check_no_run(self):
        if "run" in self.sources:
            raise PipelineItemFailure(
                "Reserved filename 'run' already exists in sources"
            )
        elif "run" in self.extras:
            raise PipelineItemFailure(
                "Reserved filename 'run' already exists in extras"
            )
        elif "run" in self.inputs:
            raise RuntimeError(
                "'run' is contained in inputs, but not sources or extras"
            )


class BuildScript(BuildStrategy):
    @classmethod
    def applicable_on_directory(cls, build: "BuildSection", directory: str) -> bool:
        return False

    def _build(self) -> str:
        assert len(self.sources) == 1
        return self._build_script(self.sources[0])

    def _build_script(self, program: str) -> str:
        interpreter = self._load_shebang(program)
        self._check_tool(interpreter)
        st = self._stat(program)
        self._chmod(program, st.st_mode | 0o111)
        return program


class BuildBinary(BuildStrategy):
    @classmethod
    def applicable_on_directory(cls, build: "BuildSection", directory: str) -> bool:
        return False


class Python(BuildScript):
    name = BuildStrategyName.python
    extra_sources: Optional[str] = "extra_sources_py"

    @classmethod
    def applicable_on_files(cls, build: "BuildSection", sources: list[str]) -> bool:
        if not cls._all_end_with(sources, [".py"]):
            return False
        return True

    def _build(self):
        entrypoint = self._get_entrypoint(".py")
        if len(self.sources) == 1:
            return self._build_script(entrypoint)
        else:
            self._check_no_run()
            self._symlink(self._build_script(entrypoint), "run")
            return "."


class Shell(BuildScript):
    name = BuildStrategyName.shell

    @classmethod
    def applicable_on_files(cls, build: "BuildSection", sources: list[str]) -> bool:
        return len(sources) == 1 and sources[0].endswith(".sh")


class C(BuildBinary):
    name = BuildStrategyName.c
    extra_sources: Optional[str] = "extra_sources_c"
    extra_nonsources: Optional[str] = "headers_c"

    @classmethod
    def applicable_on_files(cls, build: "BuildSection", sources: list[str]) -> bool:
        return cls._all_end_with(sources, [".h", ".c"])

    def _build(self) -> str:
        c_flags = ["-std=c17", "-O2", "-Wall", "-lm", "-Wshadow", "-Wno-sign-compare"]
        c_flags.append(
            "-fdiagnostics-color=" + ("never" if self._env.no_colors else "always")
        )

        self._run_subprocess(
            ["gcc", *self.sources, "-o", self.target, "-I."]
            + c_flags
            + self._build_section.comp_args,
            self._build_section.program_name,
        )
        return self.target


class Cpp(BuildBinary):
    name = BuildStrategyName.cpp
    extra_sources: Optional[str] = "extra_sources_cpp"
    extra_nonsources: Optional[str] = "headers_cpp"

    @classmethod
    def applicable_on_files(cls, build: "BuildSection", sources: list[str]) -> bool:
        return cls._all_end_with(sources, [".h", ".hpp", ".cpp", ".cc"])

    def _build(self) -> str:
        cpp_flags = [
            "-std=c++20",
            "-O2",
            "-Wall",
            "-lm",
            "-Wshadow",
            "-Wno-sign-compare",
        ]
        cpp_flags.append(
            "-fdiagnostics-color=" + ("never" if self._env.no_colors else "always")
        )

        self._run_subprocess(
            ["g++", *self.sources, "-o", self.target, "-I."]
            + cpp_flags
            + self._build_section.comp_args,
            self._build_section.program_name,
        )
        return self.target


class Pascal(BuildBinary):
    name = BuildStrategyName.pascal

    @classmethod
    def applicable_on_files(cls, build: "BuildSection", sources: list[str]) -> bool:
        return cls._all_end_with(sources, [".pas"])

    def _build(self) -> str:
        pas_flags = ["-gl", "-O3", "-Sg", "-o" + self.target, "-FE."]
        self._run_subprocess(
            ["fpc"] + pas_flags + self.sources + self._build_section.comp_args,
            self._build_section.program_name,
        )
        return self.target


class Java(BuildStrategy):
    name = BuildStrategyName.java
    extra_sources: Optional[str] = "extra_sources_java"

    @classmethod
    def applicable_on_files(cls, build: "BuildSection", sources: list[str]) -> bool:
        return cls._all_end_with(sources, [".java"])

    @classmethod
    def applicable_on_directory(cls, build: "BuildSection", directory: str) -> bool:
        return False

    def _build(self):
        self._check_tool("java")
        self._check_tool("javac")
        self._check_tool("/usr/bin/bash")

        entry_class = self._get_entrypoint(".java").rstrip(".java")
        arguments = ["javac", "-d", self.target] + self.sources
        self._run_subprocess(arguments, self._build_section.program_name)
        self._check_no_run()
        run_path = os.path.join(self.target, "run")
        with self._open(run_path, "w") as run_file:
            run_file.write(
                "#!/usr/bin/bash\n"
                + f"exec java --class-path ${{0%/run}} {entry_class} $@\n"
            )
        st = self._stat(run_path)
        self._chmod(run_path, st.st_mode | 0o111)
        return self.target


class Make(BuildStrategy):
    name = BuildStrategyName.make
    _target_subdir: str = "target"

    @classmethod
    def applicable_on_files(cls, build: "BuildSection", sources: list[str]) -> bool:
        return False

    @classmethod
    def applicable_on_directory(cls, build: "BuildSection", directory: str) -> bool:
        return os.path.exists(os.path.join(directory, "Makefile"))

    def _build(self) -> str:
        directory = self._listdir()[0]
        with FakeChangedCWD(self, directory):
            if self._exists(self._target_subdir):
                raise PipelineItemFailure(
                    f"Makefile strategy: '{self._target_subdir}' already exists"
                )
            self._makedirs(self._target_subdir)
            self._run_subprocess(["make"], self._build_section.program_name)
            if not self._isdir(self._target_subdir):
                raise PipelineItemFailure(
                    f"Makefile must create '{self._target_subdir}/' directory"
                )
        return os.path.join(directory, self._target_subdir)


class Cargo(BuildStrategy):
    name = BuildStrategyName.cargo
    _target_subdir: str = "target"
    _artifact_dir: str = ".pisek-executables"

    @classmethod
    def applicable_on_files(cls, build: "BuildSection", sources: list[str]) -> bool:
        return False

    @classmethod
    def applicable_on_directory(cls, build: "BuildSection", directory: str) -> bool:
        return os.path.exists(os.path.join(directory, "Cargo.toml"))

    def _build(self) -> str:
        directory = self._listdir()[0]
        with FakeChangedCWD(self, directory):
            if self._exists(self._target_subdir):
                raise PipelineItemFailure(
                    f"Cargo strategy: '{self._target_subdir}' already exists"
                )

            args = [
                "--release",
                "--workspace",
                "--bins",
                "--quiet",
                "--color",
                ("never" if self._env.no_colors else "always"),
            ]

            self._run_subprocess(
                [
                    "cargo",
                    "check",
                    *args,
                ],
                self._build_section.program_name,
            )

            output = self._run_subprocess(
                [
                    "cargo",
                    "build",
                    *args,
                    "--message-format",
                    "json",
                ],
                self._build_section.program_name,
            )

        self._makedirs(self._artifact_dir)
        exectables = []

        for line in output.splitlines():
            content = json.loads(line)

            if content["reason"] != "compiler-artifact":
                continue
            if "bin" not in content["target"]["kind"]:
                continue
            path = content["executable"]

            name = os.path.basename(path)
            exectables.append(os.path.basename(name))

            self._copy(path, os.path.join(self._artifact_dir, name))

        if len(exectables) == 1 and exectables != ["run"]:
            self._symlink(
                exectables[0],
                os.path.join(self._artifact_dir, "run"),
            )

        return self._artifact_dir


AUTO_STRATEGIES: list[type[BuildStrategy]] = [
    Python,
    Shell,
    C,
    Cpp,
    Pascal,
    Java,
    Make,
    Cargo,
]



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import hashlib
import time
from typing import Any, Iterable
import os
import pickle

from pisek.version import __version__
from pisek.utils.text import eprint
from pisek.utils.colors import ColorSettings
from pisek.utils.paths import INTERNALS_DIR


CACHE_VERSION_FILE = os.path.join(INTERNALS_DIR, "cache_version")
CACHE_CONTENT_FILE = os.path.join(INTERNALS_DIR, "cache")
HASH_INDEX_FILE = os.path.join(INTERNALS_DIR, "hash_index")
SAVED_LAST_SIGNATURES = 5
CACHE_SAVE_INTERVAL = 1  # seconds


class CacheEntry:
    """Object representing single cached job."""

    def __init__(
        self,
        name: str,
        signature: str,
        result: Any,
        envs: Iterable[tuple[str, ...]],
        files: Iterable[str],
        globs: Iterable[str],
        prerequisites_results: Iterable[str],
        output: list[tuple[str, bool]],
        logs: list[tuple[str, str]],
    ) -> None:
        self.name = name
        self.signature = signature
        self.result = result
        self.prerequisites_results = list(sorted(prerequisites_results))
        self.envs = list(sorted(envs))
        self.files = list(sorted(files))
        self.globs = list(sorted(globs))
        self.output = output
        self.logs = logs

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(name={self.name}, signature={self.signature}, "
            f"result={self.result}, prerequisites_results={self.prerequisites_results}, "
            f"envs={self.envs}, files={self.files}, globs={self.globs}, output={self.output},"
            f"logs={self.logs})"
        )


class Cache:
    """Object for caching jobs and file hashes."""

    def __init__(self) -> None:
        os.makedirs(INTERNALS_DIR, exist_ok=True)
        with open(CACHE_VERSION_FILE, "w") as f:
            f.write(f"{__version__}\n")
        self.cache: dict[str, list[CacheEntry]] = {}
        self.hash_index: dict[str, tuple[float, str]] = {}
        self.last_save = time.time()

    @classmethod
    def load(cls) -> "Cache":
        """Load cache file."""
        CACHE_FILES = [CACHE_VERSION_FILE, CACHE_CONTENT_FILE, HASH_INDEX_FILE]
        cache_existence = list(map(os.path.exists, CACHE_FILES))

        if not all(cache_existence):
            if any(cache_existence):
                eprint(
                    ColorSettings.colored(
                        "Incomplete cache found. Starting from scratch...",
                        "yellow",
                    )
                )
            return Cache()

        with open(CACHE_VERSION_FILE) as f:
            version = f.read().strip()

        if version != __version__:
            eprint(
                ColorSettings.colored(
                    "Different version of cache found. Starting from scratch...",
                    "yellow",
                )
            )
            return Cache()

        cache = Cache()
        cache.cache = cls.pickle_load(CACHE_CONTENT_FILE)
        cache.hash_index = cls.pickle_load(HASH_INDEX_FILE)
        return cache

    def add(self, cache_entry: CacheEntry):
        """Add entry to cache."""
        if cache_entry.name not in self.cache:
            self.cache[cache_entry.name] = []
        self.cache[cache_entry.name].append(cache_entry)

        # trim number of entries per cache name in order to limit cache size
        self.cache[cache_entry.name] = self.cache[cache_entry.name][
            -SAVED_LAST_SIGNATURES:
        ]

        # Throttling saves time massively
        if time.time() - self.last_save > CACHE_SAVE_INTERVAL:
            self.export()

    def __contains__(self, name: str) -> bool:
        return name in self.cache

    def __getitem__(self, name: str) -> list[CacheEntry]:
        return self.cache[name]

    def entry_names(self) -> list[str]:
        return list(self.cache.keys())

    def last_entry(self, name: str) -> CacheEntry:
        return self[name][-1]

    def move_to_top(self, entry: CacheEntry):
        """Move given entry to most recent position."""
        if entry in self.cache[entry.name]:
            self.cache[entry.name].remove(entry)
            self.cache[entry.name].append(entry)
        else:
            raise ValueError(
                f"Cannot move to top entry which is not in Cache:\n{entry}"
            )

    @staticmethod
    def pickle_load(path: str) -> Any:
        with open(path, "rb") as f:
            return pickle.load(f)

    @staticmethod
    def pickle_dump(obj: Any, path: str) -> None:
        """Atomic pickle.dump."""
        tmp_path = path + ".tmp"
        with open(tmp_path, "wb") as f:
            pickle.dump(obj, f)
        os.rename(tmp_path, path)

    def export(self) -> None:
        """Export cache into a file."""
        self.pickle_dump(self.cache, CACHE_CONTENT_FILE)
        self.pickle_dump(self.hash_index, HASH_INDEX_FILE)
        self.last_save = time.time()

    def file_hash(self, path: str):
        mtime = os.path.getmtime(path)
        if path not in self.hash_index or self.hash_index[path][0] != mtime:
            with open(path, "rb") as f:
                file_sign = hashlib.file_digest(f, "sha256")
            self.hash_index[path] = (mtime, file_sign.hexdigest())

        return self.hash_index[path][1]



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import json
from typing import Any, Optional

from pisek.utils.paths import TaskPath
from pisek.jobs.jobs import Job, PipelineItemFailure
from pisek.task_jobs.task_manager import (
    TaskJobManager,
    SOLUTION_MAN_CODE,
)
from pisek.task_jobs.solution.solution_result import (
    SolutionResult,
    RelativeSolutionResult,
    AbsoluteSolutionResult,
)

TESTING_LOG = "testing_log.json"


class CreateTestingLog(TaskJobManager):
    run_always: bool = True

    def __init__(self):
        super().__init__("Creating testing log")

    def _get_jobs(self) -> list[Job]:
        return []

    def _evaluate(self) -> None:
        log: dict[str, Any] = {"source": "pisek", "solutions": {}}
        solutions: set[str] = set()
        warn_skipped: bool = False
        for name, data in self.prerequisites_results.items():
            if not name.startswith(SOLUTION_MAN_CODE) or not any(
                data["results"].values()
            ):
                continue

            solution = name[len(SOLUTION_MAN_CODE) :]
            solutions.add(solution)
            log["solutions"][solution] = {"results": {}}
            solution_results = log["solutions"][solution]["results"]

            inp: TaskPath
            sol_res: Optional[SolutionResult]
            for inp, sol_res in data["results"].items():
                if sol_res is None:
                    warn_skipped = True
                    continue
                solution_results[inp.name] = {
                    "time": sol_res.solution_rr.time,
                    "wall_clock_time": sol_res.solution_rr.wall_time,
                    "result": sol_res.verdict.name,
                }

                if isinstance(sol_res, RelativeSolutionResult):
                    solution_results[inp.name]["relative_points"] = str(
                        sol_res.relative_points
                    )
                elif isinstance(sol_res, AbsoluteSolutionResult):
                    solution_results[inp.name]["absolute_points"] = str(
                        sol_res.absolute_points
                    )
                else:
                    raise ValueError(
                        f"Unknown {SolutionResult.__name__} instance found: {type(sol_res)}"
                    )

        if len(solutions) == 0:
            raise PipelineItemFailure("No solution was tested.")

        if len(solutions) < len(self._env.config.solutions):
            self._warn("Not all solutions were tested.")
        if warn_skipped:
            self._warn("Not all inputs were tested. For testing them use --all-inputs.")

        with open(TaskPath(TESTING_LOG).path, "w") as f:
            json.dump(log, f, indent=4)



# pisek cms - Tool for importing tasks from Pisek into CMS.
#
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Any
from cms.db.task import Dataset
from cms.db.submission import SubmissionResult, Evaluation
from cms.db.filecacher import FileCacher
from sqlalchemy.orm import Session
import json

from pisek.user_errors import TestingFailed
from pisek.cms.submission import get_submission
from pisek.utils.colors import ColorSettings
from pisek.env.env import Env
from pisek.task_jobs.testing_log import TESTING_LOG
from pisek.task_jobs.solution.solution_result import Verdict
from pisek.config.task_config import SolutionSection, TestSection
from pisek.utils.text import eprint, tab

RUNTIME_ERROR_MESSAGES = [
    "Evaluation didn't produce file %s",
    "Execution killed (could be triggered by violating memory limits)",
    "Execution failed because the return code was nonzero",
]

TIME_LIMIT_MESSAGES = [
    "Execution timed out",
    "Execution timed out (wall clock limit exceeded)",
]


def create_testing_log(session: Session, env: Env, dataset: Dataset) -> None:
    config = env.config
    files = FileCacher()

    payload: dict[str, Any] = {"source": "cms", "solutions": {}}
    success = True

    for name, solution in config.solutions.items():
        results: dict[str, Any] = {}
        payload["solutions"][name] = {"results": results}

        try:
            result = get_submission_result(session, files, env, solution, dataset)
        except SubmissionResultError as e:
            eprint(ColorSettings.colored(f"Skipping {name}: {e}", "yellow"))
            success = False
            continue

        evaluation: Evaluation
        for evaluation in result.evaluations:
            points: str | float
            result_type: str

            try:
                points = float(evaluation.outcome)

                if points >= 1:
                    result_type = Verdict.ok.name
                elif points <= 0:
                    if evaluation.text[0] in TIME_LIMIT_MESSAGES:
                        result_type = Verdict.timeout.name
                    elif evaluation.text[0] in RUNTIME_ERROR_MESSAGES:
                        result_type = Verdict.error.name
                    else:
                        result_type = Verdict.wrong_answer.name
                else:
                    result_type = Verdict.partial_ok.name
            except ValueError:
                points = evaluation.outcome
                result_type = "indeterminate"

            result = {
                "time": evaluation.execution_time,
                "wall_clock_time": evaluation.execution_wall_clock_time,
                "memory": (
                    evaluation.execution_memory / (1024 * 1024)
                    if evaluation.execution_memory is not None
                    else None
                ),
                "relative_points": points,
                "result": result_type,
            }

            result = dict(filter(lambda p: p[1] is not None, result.items()))

            results[f"{evaluation.codename}.in"] = result

    with open(TESTING_LOG, "w") as file:
        json.dump(payload, file, indent=4)

    if not success:
        raise TestingFailed()


def check_results(session: Session, env: Env, dataset: Dataset) -> None:
    config = env.config
    files = FileCacher()

    success = True

    solution: SolutionSection
    for name, solution in config.solutions.items():
        try:
            result = get_submission_result(session, files, env, solution, dataset)

            if not result.scored():
                raise SubmissionResultError("This submission has not been scored yet")
        except SubmissionResultError as e:
            print(ColorSettings.colored(f"Skipping {name}: {e}", "yellow"))
            success = False
            continue

        score = result.score

        score_missed_target = None

        if solution.points is not None and score != solution.points:
            score_missed_target = f"{solution.points}"
        elif solution.points_min is not None and score < solution.points_min:
            score_missed_target = f"above {solution.points_min}"
        elif solution.points_max is not None and score > solution.points_max:
            score_missed_target = f"below {solution.points_max}"

        message = f"{name}: {score} points"

        if score_missed_target is not None:
            message += f" (should be {score_missed_target})"
            message = ColorSettings.colored(message, "red")
            success = False

        print(message)

        subtasks: list[tuple[int, TestSection]] = list(config.test_sections.items())
        fractions = get_subtask_score_fractions(result.score_details)

        if fractions is None or len(fractions) != len(subtasks):
            message = "The task seems to use an unsupported score type, skipping checking subtasks"
            print(tab(ColorSettings.colored(message, "red")))

            success = False
            continue

        target: str
        for (num, subtask), fraction, target in zip(
            subtasks, fractions, solution.tests
        ):
            name = subtask.name or f"Subtask {num}"

            if target == "X":
                correct = True
            elif target == "1":
                target_name = "correct"
                correct = fraction == 1.0
            elif target == "P":
                target_name = "partially correct"
                correct = 0.0 < fraction < 1.0
            else:
                assert target in (
                    "0",
                    "W",
                    "T",
                    "!",
                ), f"Unknown expected result '{target}'"

                target_name = "wrong"
                correct = fraction == 0.0

            message = f"{name}: {fraction}"

            if not correct:
                message += f" (should be {target_name})"
                message = ColorSettings.colored(message, "red")
                success = False

            print(tab(message))

    if not success:
        raise TestingFailed()


def get_subtask_score_fractions(score_details: Any) -> Optional[list[float]]:
    if not isinstance(score_details, list):
        return None

    results = []

    for subtask in score_details:
        if not isinstance(subtask, dict):
            return None

        if "score_fraction" not in subtask:
            return None

        fraction = subtask["score_fraction"]

        if not isinstance(fraction, float):
            return None

        results.append(fraction)

    return results


def get_submission_result(
    session: Session,
    files: FileCacher,
    env: Env,
    solution: SolutionSection,
    dataset: Dataset,
) -> SubmissionResult:
    submission = get_submission(session, files, env, solution, dataset.task)

    if submission is None:
        raise SubmissionResultError("This solution has not been submitted yet")

    result: Optional[SubmissionResult] = (
        session.query(SubmissionResult)
        .filter(SubmissionResult.submission == submission)
        .filter(SubmissionResult.dataset == dataset)
        .one_or_none()
    )

    if result is None:
        raise SubmissionResultError(
            "The latest submission hasn't started evaluating on this dataset"
        )

    if result.compilation_failed():
        raise SubmissionResultError("The submission failed to compile")

    if not result.evaluated():
        raise SubmissionResultError("The submission is still being evaluated")

    return result


class SubmissionResultError(Exception):
    pass



#!/usr/bin/env python3
def read_ints(n: int):
    inp = input()
    assert inp == inp.strip()
    ints = list(map(int, inp.split(" ")))
    assert len(ints) == n
    return ints


(n,) = read_ints(1)
nums = read_ints(n)

try:
    input()
except EOFError:
    quit(42)
raise ValueError("Input continues.")



#!/usr/bin/env python3

from random import Random
import os
import sys

if len(sys.argv) == 1:
    print("random-uniform repeat=10")
    exit()

assert sys.argv[1] == "random-uniform"
assert len(sys.argv[2]) == 16
random = Random(int(sys.argv[2], base=16))

nums = set()

n = random.randint(300000, 400000)
for _ in range(n):
    nums.add(random.randint(1, 1_000_000))

nums = list(nums)
random.shuffle(nums)

for num in nums:
    print(num)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from decimal import Decimal, InvalidOperation
from typing import Optional
from tempfile import gettempdir
from uuid import uuid4

from pisek.utils.paths import IInputPath, IOutputPath
from pisek.env.env import Env
from pisek.config.task_config import RunSection
from pisek.config.config_types import ProgramRole
from pisek.task_jobs.run_result import RunResult
from pisek.task_jobs.solution.solution_result import (
    Verdict,
    SolutionResult,
    RelativeSolutionResult,
)
from pisek.task_jobs.checker.checker_base import RunChecker, RunBatchChecker


class RunCMSJudge(RunChecker):
    """Judge class with CMS helper functions"""

    def __init__(
        self,
        env: Env,
        judge: RunSection,
        **kwargs,
    ) -> None:
        super().__init__(env=env, checker_name=judge.name, **kwargs)
        self.judge = judge
        self.points_file = self.checker_log_file.replace_suffix(".points")

    def _load_points(self, result: RunResult) -> Decimal:
        with self._open_file(result.stdout_file) as f:
            points_str = f.read().split("\n")[0]
        try:
            points = Decimal(points_str)
        except (ValueError, InvalidOperation):
            raise self._create_program_failure(
                "Judge didn't write points on stdout:",
                result,
                status=False,
                stdout_force_content=True,
            )

        if not 0 <= points <= 1:
            raise self._create_program_failure(
                "Judge must give between 0 and 1 points:",
                result,
                status=False,
                stdout_force_content=True,
            )

        return points

    def _load_solution_result(self, judge_run_result: RunResult) -> SolutionResult:
        if judge_run_result.returncode == 0:
            points = self._load_points(judge_run_result)
            if points == 1.0:
                verdict = Verdict.ok
            elif points == 0.0:
                verdict = Verdict.wrong_answer
            else:
                verdict = Verdict.partial_ok

            with self._open_file(judge_run_result.stderr_file) as f:
                message = f.readline().removesuffix("\n")

            return RelativeSolutionResult(
                verdict=verdict,
                message=message,
                solution_rr=self._solution_run_res,
                checker_rr=judge_run_result,
                relative_points=points,
            )
        else:
            raise self._create_program_failure(
                f"Judge failed on {self._checking_message()}:",
                judge_run_result,
                stderr_force_content=True,
            )


class RunCMSBatchJudge(RunCMSJudge, RunBatchChecker):
    """Checks solution output using judge with CMS interface."""

    def __init__(
        self,
        env: Env,
        judge: RunSection,
        test: int,
        input_: IInputPath,
        output: IOutputPath,
        correct_output: IOutputPath,
        expected_verdict: Optional[Verdict],
        **kwargs,
    ) -> None:
        super().__init__(
            env=env,
            judge=judge,
            test=test,
            input_=input_,
            output=output,
            correct_output=correct_output,
            expected_verdict=expected_verdict,
            **kwargs,
        )

    @staticmethod
    def _invalid_path(name: str):
        return os.path.join(gettempdir(), f"the-{name}-is-not-available-{uuid4()}")

    def _check(self) -> SolutionResult:
        config = self._env.config

        self._access_file(self.output)
        if config.tests.judge_needs_in:
            self._access_file(self.input)
        if config.tests.judge_needs_out:
            self._access_file(self.correct_output)

        result = self._run_program(
            ProgramRole.judge,
            self.judge,
            args=[
                (
                    self.input.abspath
                    if config.tests.judge_needs_in
                    else RunCMSBatchJudge._invalid_path("input")
                ),
                (
                    self.correct_output.abspath
                    if config.tests.judge_needs_out
                    else RunCMSBatchJudge._invalid_path("output")
                ),
                self.output.abspath,
            ],
            stdout=self.points_file,
            stderr=self.checker_log_file,
        )

        sol_result = self._load_solution_result(result)
        return sol_result



#!/usr/bin/env python3
import sys
from typing import NoReturn


def award(points: float, msg: str) -> NoReturn:
    print(points)
    print(msg, file=sys.stderr)
    exit(0)


def reject() -> NoReturn:
    award(0, "translate:wrong")


input_, correct_output, contestant_output = sys.argv[1:]

with open(input_) as f:
    n, k = map(int, f.readline().split())

with open(contestant_output) as f:
    try:
        numbers = list(map(int, f.readline().split()))
    except ValueError:
        reject()  # The contestant did not print integers
    except EOFError:
        reject()  # The contestant output ends

    if f.read().strip():
        reject()  # Contestant output doesn't end when it should


# Be careful to check **ALL** constraints
if any(map(lambda x: x <= 0, numbers)):
    reject()

if sum(numbers) == k:
    award(1.0, "translate:success")
elif sum(numbers) >= k / 2:
    award(0.5, "translate:partial")
else:
    reject()  # Not enough sand used



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argcomplete
import argparse
import logging
import os
import sys
from typing import Optional

from pisek.user_errors import (
    UserError,
    TaskConfigError,
    TestingFailed,
    MissingFile,
    InvalidArgument,
    InvalidOperation,
)

from pisek.utils.util import clean_task_dir, log_level_mapper
from pisek.utils.text import eprint
from pisek.utils.colors import ColorSettings

from pisek.visualize import visualize
from pisek.init import init_task
from pisek.config.config_hierarchy import DEFAULT_CONFIG_FILENAME
from pisek.config.config_tools import export_config, update_and_replace_config
from pisek.version import print_version

from pisek.jobs.task_pipeline import TaskPipeline
from pisek.utils.pipeline_tools import (
    run_pipeline,
    PATH,
    locked_folder,
    assert_task_dir,
)
from pisek.utils.paths import INTERNALS_DIR

LOG_FILE = os.path.join(INTERNALS_DIR, "log")


@locked_folder
def test_task(args, **kwargs) -> None:
    return test_task_path(PATH, **vars(args), **kwargs)


def test_task_path(path, solutions: Optional[list[str]] = None, **env_args) -> None:
    return run_pipeline(path, TaskPipeline, solutions=solutions, **env_args)


def test_solutions(args) -> None:
    return test_task(args)


def test_generator(args) -> None:
    return test_task(args, solutions=[])


@locked_folder
def clean_directory() -> None:
    task_dir = PATH
    eprint(f"Cleaning directory: {os.path.abspath(task_dir)}")
    return clean_task_dir(task_dir)


def _main(argv: list[str]) -> None:
    parser = argparse.ArgumentParser(
        description=(
            "Tool for developing tasks for programming competitions. "
            "Full documentation is at https://github.com/kasiopea-org/pisek"
        )
    )

    def add_cms_import_arguments(parser):
        parser.add_argument(
            "--description",
            "-d",
            type=str,
            help="create the dataset with the description DESCRIPTION",
        )

        parser.add_argument(
            "--time-limit",
            "-t",
            type=float,
            help="override the time limit when importing to TIME_LIMIT seconds",
        )

    def add_argument_dataset(parser):
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument(
            "--dataset",
            "-d",
            type=str,
            help="use the dataset with the description DESCRIPTION",
        )

        group.add_argument(
            "--active-dataset",
            "-a",
            action="store_true",
            help="use active dataset",
        )

    # ------------------------------- pisek -------------------------------

    parser.add_argument(
        "--clean",
        "-c",
        action="store_true",
        help="clean directory beforehand",
    )
    parser.add_argument(
        "--plain",
        "-p",
        action="store_true",
        help="do not use ANSI escape sequences",
    )
    parser.add_argument(
        "--no-jumps",
        action="store_true",
        help="do not use ANSI cursor movement & clear sequences",
    )
    parser.add_argument(
        "--no-colors",
        action="store_true",
        help="do not use ANSI color sequences",
    )
    parser.add_argument(
        "--pisek-dir",
        help="pisek directory for higher level settings",
        type=str,
    )
    parser.add_argument(
        "--config-filename",
        help="override name of the configuration file",
        default=DEFAULT_CONFIG_FILENAME,
        type=str,
    )

    subparsers = parser.add_subparsers(
        help="subcommand to run", dest="subcommand", required=True
    )

    # ------------------------------- pisek version -------------------------------

    parser_version = subparsers.add_parser("version", help="print current version")

    # ------------------------------- pisek test -------------------------------

    parser_test = subparsers.add_parser("test", help="test task")
    test_subparsers = parser_test.add_subparsers(help="testing target", dest="target")
    test_all = test_subparsers.add_parser("all", help="test all")
    test_gen = test_subparsers.add_parser("generator", help="test only generator")
    test_sols = test_subparsers.add_parser(
        "solutions", help="test generator & given solutions"
    )
    test_sols.add_argument(
        "solutions", type=str, help="name of the solutions to test", nargs="+"
    )

    parser_test.add_argument(
        "--jobs",
        "-j",
        type=int,
        help="how many jobs to run in parallel",
    )
    parser_test.add_argument(
        "--verbosity",
        "-v",
        action="count",
        default=0,
        help="be more verbose (enter multiple times for even more verbosity)",
    )
    parser_test.add_argument(
        "--file-contents",
        "-C",
        action="store_true",
        help="show file contents on error",
    )
    parser_test.add_argument(
        "--time-limit",
        "-t",
        type=float,
        help="override time limit for solutions to TIME_LIMIT seconds",
    )
    parser_test.add_argument(
        "--full", "-f", action="store_true", help="don't stop on first failure"
    )
    parser_test.add_argument(
        "--strict",
        action="store_true",
        help="interpret warnings as failures (for final check)",
    )
    parser_test.add_argument(
        "--repeat",
        "-n",
        type=int,
        default=1,
        help="test task REPEAT times giving generator different seeds. (Changes seeded inputs only.)",
    )
    parser_test.add_argument(
        "--all-inputs",
        "-a",
        action="store_true",
        help="test each solution on all inputs",
    )
    parser_test.add_argument(
        "--testing-log",
        "-T",
        action="store_true",
        help="write test results to testing_log.json",
    )

    # ------------------------------- pisek clean -------------------------------

    parser_clean = subparsers.add_parser("clean", help="clean task directory")

    # ------------------------------- pisek init -------------------------------

    parser_task = subparsers.add_parser("init", help="create a task skeleton")

    # ------------------------------- pisek config -------------------------------

    parser_config = subparsers.add_parser("config", help="manage task config")
    config_subparsers = parser_config.add_subparsers(
        help="subcommand to run", dest="config_subcommand", required=True
    )
    config_subparsers.add_parser(
        "update", help="update config to newest version (replaces the config)"
    )
    parser_export = config_subparsers.add_parser(
        "export", help="creates an organization independent config"
    )
    parser_export.add_argument(
        "filename", help="name of the exported configuration file"
    )

    # ------------------------------- pisek visualize -------------------------------

    parser_visualize = subparsers.add_parser(
        "visualize", help="show solution statistics and closeness to limit"
    )
    parser_visualize.add_argument(
        "--filter",
        "-f",
        choices=("slowest", "all"),
        default="slowest",
        type=str,
        help="which inputs to show (slowest/all)",
    )
    parser_visualize.add_argument(
        "--bundle",
        "-b",
        action="store_true",
        help="don't group inputs by test",
    )
    parser_visualize.add_argument(
        "--solutions",
        "-s",
        default=None,
        type=str,
        nargs="*",
        help="visualize only solutions with a name or source in SOLUTIONS",
    )
    parser_visualize.add_argument(
        "--filename",
        default="testing_log.json",
        type=str,
        help="read testing log from FILENAME",
    )
    parser_visualize.add_argument(
        "--limit",
        "-l",
        default=None,
        type=float,
        help="visualize as if the time limit was LIMIT seconds",
    )
    parser_visualize.add_argument(
        "--segments",
        "-S",
        type=int,
        help="print bars SEGMENTS characters wide",
    )

    # ------------------------------- pisek cms -------------------------------

    parser_cms = subparsers.add_parser("cms", help="import task into CMS")
    subparsers_cms = parser_cms.add_subparsers(
        help="subcommand to run", dest="cms_subcommand", required=True
    )

    parser_cms_create = subparsers_cms.add_parser("create", help="create a new task")
    add_cms_import_arguments(parser_cms_create)

    parser_cms_update = subparsers_cms.add_parser(
        "update", help="update the basic properties of an existing task"
    )

    parser_cms_add = subparsers_cms.add_parser(
        "add", help="add a dataset to an existing task"
    )
    add_cms_import_arguments(parser_cms_add)
    parser_cms_add.add_argument(
        "--no-autojudge",
        action="store_true",
        help="disable background judging for the new dataset",
    )

    parser_cms_submit = subparsers_cms.add_parser(
        "submit", help="submit reference solutions for evaluation using CMS"
    )
    parser_cms_submit.add_argument(
        "--username",
        "-u",
        type=str,
        required=True,
        help="submit as the user with username USERNAME",
    )

    parser_cms_testing_log = subparsers_cms.add_parser(
        "testing-log",
        help="generate a testing log for reference solutions submitted to CMS",
    )
    add_argument_dataset(parser_cms_testing_log)

    parser_cms_check = subparsers_cms.add_parser(
        "check",
        help="check if reference solutions scored as expected in CMS",
    )
    add_argument_dataset(parser_cms_check)

    argcomplete.autocomplete(parser)
    args = parser.parse_args(argv)

    if args.pisek_dir is None and "PISEK_DIRECTORY" in os.environ:
        args.pisek_dir = os.path.join(os.getcwd(), os.environ["PISEK_DIRECTORY"])
    ColorSettings.set_state(not args.plain and not args.no_colors)

    # Taskless subcommands
    if args.subcommand == "version":
        return print_version()
    elif args.subcommand == "init":
        return init_task(args.config_filename, args.no_jumps)

    # !!! Ensure this is always run before clean_directory !!!
    assert_task_dir(PATH, args.pisek_dir, args.config_filename)

    if args.clean:
        clean_directory()

    os.makedirs(os.path.dirname(LOG_FILE), exist_ok=True)
    open(LOG_FILE, "w").close()
    logging.basicConfig(
        filename=LOG_FILE,
        encoding="utf-8",
        level=log_level_mapper(os.getenv("LOG_LEVEL", "info")),
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
    )

    if args.subcommand == "test":
        if args.target == "generator":
            return test_generator(args)
        elif args.target == "solutions":
            return test_solutions(args)
        elif args.target is None or args.target == "all":
            return test_task(args, solutions=None)
        else:
            assert False, "Unknown command"

    elif args.subcommand == "config":
        if args.config_subcommand == "update":
            return update_and_replace_config(PATH, args.pisek_dir, args.config_filename)
        elif args.config_subcommand == "export":
            return export_config(
                PATH,
                args.pisek_dir,
                args.config_filename,
                os.path.join(PATH, args.filename),
            )
        else:
            assert False, "Unknown command"

    elif args.subcommand == "cms":
        args, unknown_args = parser.parse_known_args()

        try:
            import pisek.cms as cms
        except ImportError as err:
            err.add_note("Failed to locate CMS installation")
            raise

        from logging import StreamHandler
        from sys import stdout

        root_logger = logging.getLogger()
        for handler in root_logger.handlers:
            if isinstance(handler, StreamHandler) and handler.stream is stdout:
                root_logger.removeHandler(handler)

        if args.cms_subcommand == "create":
            return cms.create(args)
        elif args.cms_subcommand == "update":
            return cms.update(args)
        elif args.cms_subcommand == "add":
            return cms.add(args)
        elif args.cms_subcommand == "submit":
            return cms.submit(args)
        elif args.cms_subcommand == "testing-log":
            return cms.testing_log(args)
        elif args.cms_subcommand == "check":
            return cms.check(args)
        else:
            assert False, "Unknown command"

    elif args.subcommand == "clean":
        return clean_directory()
    elif args.subcommand == "visualize":
        return visualize(PATH, **vars(args))
    else:
        assert False, "Unknown command"


def main(argv: list[str]) -> int:
    try:
        _main(argv)
    except TestingFailed:
        return 1
    except (TaskConfigError, MissingFile, InvalidArgument, InvalidOperation) as e:
        print(ColorSettings.colored(str(e), "red"))
        return 2
    except UserError:
        raise NotImplementedError()
    except KeyboardInterrupt:
        eprint("\rStopping...")
        return 130

    return 0


def main_wrapped() -> int:
    return main(sys.argv[1:])


if __name__ == "__main__":
    exit(main_wrapped())



import os

import unittest

from pisek.utils.paths import TESTS_DIR, GENERATED_SUBDIR
from util import TestFixtureVariant, overwrite_file, modify_config


class TestSumCMS(TestFixtureVariant):
    @property
    def fixture_path(self) -> str:
        return "fixtures/sum_cms/"


class TestMissingGenerator(TestSumCMS):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        os.remove(os.path.join(self.task_dir, "gen.py"))


class TestGeneratorDoesNotCreateTests(TestSumCMS):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        overwrite_file(self.task_dir, "gen.py", "gen_dummy.py")


class TestMissingInputFilesForTest(TestSumCMS):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        overwrite_file(self.task_dir, "gen.py", "gen_incomplete.py")


class TestOldInputsDeleted(TestSumCMS):
    """Do we get rid of out-of-date inputs?"""

    def modify_task(self) -> None:
        self.inputs_dir = os.path.join(self.task_dir, TESTS_DIR, GENERATED_SUBDIR)

        os.makedirs(os.path.join(self.inputs_dir), exist_ok=True)

        with open(os.path.join(self.inputs_dir, "01_outdated.in"), "w") as f:
            # This old input does not conform to the test! Get rid of it.
            f.write("-3 -2\n")

    def check_end_state(self) -> None:
        self.assertNotIn("01_outdated.in", os.listdir(self.inputs_dir))


class TestDifferentlyScoringSolution(TestSumCMS):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        overwrite_file(
            self.task_dir, "solve_0b.py", "solve_3b.cpp", new_file_name="solve_0b.cpp"
        )


class TestNoSolutions(TestSumCMS):
    def modify_task(self) -> None:
        def modification_fn(raw_config):
            for section in raw_config.sections():
                if section.startswith("solution_"):
                    del raw_config[section]

        modify_config(self.task_dir, modification_fn)


class TestPartialJudge(TestSumCMS):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        overwrite_file(self.task_dir, "judge.cpp", "judge_no_partial.cpp")


class TestInvalidJudgeScore(TestSumCMS):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        overwrite_file(self.task_dir, "judge.cpp", "judge_invalid_score.cpp")


class TestStrictValidator(TestSumCMS):
    """A validator whose bounds are stricter than what the generator creates."""

    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        overwrite_file(self.task_dir, "validate.py", "validate_strict.py")


class TestDirtySample(TestSumCMS):
    """Sample without newline at the end."""

    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        with open(os.path.join(self.task_dir, "sample_01.in"), "w") as f:
            f.write("1 2")


class TestNoLFInStrictTextInput(TestSumCMS):
    """Input without newline at the end with in_format=text."""

    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["in_gen"] = "gen_no_lf"
            raw_config["tests"].pop("validator")
            raw_config["tests"].pop("validator_type")

        modify_config(self.task_dir, modification_fn)


class TestNoLFInTextInput(TestSumCMS):
    """Input without newline at the end with in_format=text."""

    def expecting_success(self) -> bool:
        return True

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["in_format"] = "text"
            raw_config["tests"]["in_gen"] = "gen_no_lf"
            raw_config["tests"].pop("validator")
            raw_config["tests"].pop("validator_type")

        modify_config(self.task_dir, modification_fn)


class TestNoLFInBinaryInput(TestSumCMS):
    """Input without newline at the end with in_format=binary."""

    def expecting_success(self) -> bool:
        return True

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["in_format"] = "binary"
            raw_config["tests"]["in_gen"] = "gen_no_lf"
            raw_config["tests"].pop("validator")
            raw_config["tests"].pop("validator_type")

        modify_config(self.task_dir, modification_fn)


class TestNoLFInTextOutput(TestSumCMS):
    """Output without newline at the end with out_format=text."""

    def expecting_success(self) -> bool:
        return True

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["out_format"] = "text"
            raw_config["solution_solve_no_lf"]["tests"] = "1111"

        modify_config(self.task_dir, modification_fn)


class TestNoLFInBinaryOutput(TestSumCMS):
    """Output without newline at the end with out_format=binary."""

    def expecting_success(self) -> bool:
        return True

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["out_format"] = "binary"
            raw_config["solution_solve_no_lf"]["tests"] = "1111"

        modify_config(self.task_dir, modification_fn)


class TestRustJudge(TestSumCMS):
    """A judge written in Rust."""

    def expecting_success(self) -> bool:
        return True

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["out_judge"] = "judge_rust"
            raw_config["tests"]["judge_needs_out"] = "no"

        modify_config(self.task_dir, modification_fn)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from configparser import ConfigParser
from copy import copy
import glob
from itertools import product
import os
import re

from pisek.user_errors import TaskConfigError
from pisek.utils.text import eprint
from pisek.utils.colors import ColorSettings
from pisek.config.config_types import ProgramRole


def maybe_rename_key(
    config: ConfigParser, section_from: str, section_to: str, key_from: str, key_to: str
):
    if section_from in config and key_from in config[section_from]:
        if section_to not in config:
            config.add_section(section_to)

        config[section_to][key_to] = config[section_from][key_from]
        del config[section_from][key_from]


def maybe_move_key(config: ConfigParser, key: str, section_from: str, section_to: str):
    maybe_rename_key(config, section_from, section_to, key, key)


def maybe_delete_key(config: ConfigParser, section: str, key: str):
    if section in config and key in config[section]:
        del config[section][key]


def update_to_v2(config: ConfigParser, task_path: str) -> None:
    maybe_rename_key(config, "task", "task", "samples_subdir", "static_subdir")
    maybe_rename_key(config, "tests", "tests", "solution_manager", "stub")

    subtask_points = []
    for section in sorted(config.sections()):
        if re.fullmatch(r"test[0-9]{2}", section):
            if "points" not in config[section]:
                raise TaskConfigError(f"Missing key 'points' in section [{section}]")
            subtask_points.append(int(config[section]["points"]))
    if "test00" not in config.sections():
        subtask_points = [0] + subtask_points

    if "solutions" not in config["task"]:
        raise TaskConfigError("Missing key 'solutions' in section [task]")
    solutions = config["task"]["solutions"].split()
    del config["task"]["solutions"]

    for i, solution in enumerate(solutions):
        if match := re.fullmatch(r"(.*?)_([0-9]{1,3}|X)b", solution):
            points = None if match[2] == "X" else int(match[2])
            if len(
                glob.glob(
                    os.path.join(
                        task_path,
                        config["task"].get("solutions_subdir", ""),
                        f"{solution}.*",
                    )
                )
            ):
                source = solution
            else:
                source = match[1]
        else:
            source = solution
            points = sum(subtask_points)

        if points is None:
            subtasks = "X" * len(subtask_points)
        else:
            subtasks = get_subtask_mask(points, subtask_points)

        solution = f"solution_{solution}"
        config.add_section(solution)
        if i == 0:
            config[solution]["primary"] = "yes"
            subtasks = "1" * len(subtask_points)
        config[solution]["source"] = source
        config[solution]["points"] = "X" if points is None else str(points)
        config[solution]["subtasks"] = subtasks

    subtask_inputs = {}
    in_globs_used = False
    for section in config.sections():
        if not (mat := re.fullmatch(r"test([0-9]{2})", section)):
            continue

        num = int(mat[1])
        if "in_globs" not in config[section]:
            if num == 0:
                subtask_inputs[num] = {"sample*.in"}
            else:
                subtask_inputs[num] = {f"{i:02}*.in" for i in range(1, num + 1)}
        else:
            in_globs_used = True
            subtask_inputs[num] = set(
                map(lambda x: x.replace("_*", "*"), config[section]["in_globs"].split())
            )

    last_subtask = max(subtask_inputs.keys())
    for subtask, inputs in subtask_inputs.items():
        if "0*.in" in inputs or "*.in" in inputs:
            subtask_inputs[subtask] = {
                f"{i:02}*.in" for i in range(1, last_subtask + 1)
            }

    # Now we construct ordering of subtask difficulty but without redundant edges

    # First construct edges
    subtask_includes: dict[int, set[int]] = {i: set([]) for i in subtask_inputs}
    for succ_subtask in subtask_inputs:
        for pred_subtask in subtask_inputs:
            if succ_subtask == pred_subtask:
                continue
            if subtask_inputs[succ_subtask] >= subtask_inputs[pred_subtask]:
                subtask_includes[succ_subtask].add(pred_subtask)

    # Then remove redundant edges
    for succ_subtask in subtask_includes:
        for pred_subtask in subtask_includes:
            if succ_subtask == pred_subtask:
                continue
            if pred_subtask in subtask_includes[succ_subtask]:
                subtask_includes[succ_subtask] -= subtask_includes[pred_subtask]

    if in_globs_used:
        for subtask in subtask_includes:
            subtask_section = config[f"test{subtask:02}"]
            subtask_section["predecessors"] = " ".join(
                map(str, subtask_includes[subtask])
            )

            in_globs = copy(subtask_inputs[subtask])
            for pred in subtask_includes[subtask]:
                in_globs -= subtask_inputs[pred]
            subtask_section["in_globs"] = " ".join(sorted(in_globs))


def get_subtask_mask(points, subtasks):
    all_valid = [0] * len(subtasks)

    valid = 0
    for comb in product([0, 1], repeat=len(subtasks)):
        p = sum([comb[i] * subtasks[i] for i in range(len(subtasks))])
        if p == points:
            valid += 1
            for i in range(len(subtasks)):
                all_valid[i] += comb[i]

    if valid == 0:
        return "X" * len(subtasks)

    sub_mask = ""
    for x in all_valid:
        if x == valid:
            sub_mask += "1"
        elif x == 0:
            sub_mask += "0"
        else:
            sub_mask += "X"
    return sub_mask


def update_to_extra(
    config: ConfigParser, key_from: str, key_to: str, suffix: str
) -> None:
    if "build_solution" not in config:
        config.add_section("build_solution")
    if key_from in config["tests"]:
        file = config["tests"][key_from]
        if os.path.exists(file + suffix):
            config["build_solution"][key_to] = file + suffix
        elif os.path.exists(file):
            config["build_solution"][key_to] = file


def update_to_v3(config: ConfigParser, task_path: str) -> None:
    contest_type = config.get("task", "contest_type", fallback="kasiopea")
    if contest_type not in ["kasiopea", "cms"]:
        raise TaskConfigError(f"Invalid contest_type: '{contest_type}'")
    config["task"]["use"] = f"@{contest_type}"
    maybe_delete_key(config, "task", "contest_type")

    maybe_delete_key(config, "task", "data_subdir")

    task_type = config.get("task", "task_type", fallback="batch")
    out_check = config.get("tests", "out_check")
    if contest_type == "cms" and out_check == "judge":
        if task_type == "batch":
            config["tests"]["judge_type"] = "cms-batch"
        elif task_type == "communication":
            config["tests"]["judge_type"] = "cms-communication"

    if task_type == "communication":
        config["task"]["task_type"] = "interactive"

    maybe_rename_key(config, "tests", "tests", "checker", "validator")

    OLD_NAME = {
        ProgramRole.gen: "in_gen",
        ProgramRole.validator: "checker",
        ProgramRole.primary_solution: "solve",
        ProgramRole.secondary_solution: "sec_solve",
        ProgramRole.judge: "judge",
    }

    for new_program_role in ProgramRole:
        old_program_role = OLD_NAME[new_program_role]
        if new_program_role == ProgramRole.primary_solution:
            section = "run_solution"
        else:
            section = f"run_{new_program_role}"

        config.add_section(section)
        config[section]["clock_mul"] = "0"
        config[section]["clock_min"] = config.get(
            "limits", f"{old_program_role}_clock_limit", fallback="360"
        )
        maybe_delete_key(config, "limits", f"{old_program_role}_clock_limit")

        for limit in ("time", "mem", "process"):
            maybe_rename_key(
                config,
                "limits",
                section,
                f"{old_program_role}_{limit}_limit",
                f"{limit}_limit",
            )

    update_to_extra(config, "stub", "extra_sources_c", ".c")
    update_to_extra(config, "stub", "extra_sources_cpp", ".cpp")
    update_to_extra(config, "stub", "extra_sources_py", ".py")
    update_to_extra(config, "stub", "extra_sources_java", ".java")
    maybe_delete_key(config, "tests", "stub")
    update_to_extra(config, "headers", "headers_c", ".h")
    update_to_extra(config, "headers", "headers_cpp", ".h")
    maybe_delete_key(config, "tests", "headers")

    maybe_move_key(config, "static_subdir", "task", "tests")

    if contest_type == "cms":
        maybe_move_key(config, "name", "task", "cms")
    elif contest_type == "kasiopea":
        maybe_delete_key(config, "task", "name")

    IGNORED_KEYS = [
        ("task", "tests"),
        ("tests", "in_mode"),
        ("tests", "out_mode"),
        ("tests", "online_validity"),
    ]
    for section, key in IGNORED_KEYS:
        maybe_delete_key(config, section, key)

    maybe_rename_key(
        config,
        "checks",
        "checks",
        "solution_for_each_subtask",
        "solution_for_each_test",
    )
    maybe_rename_key(
        config,
        "checks",
        "checks",
        "all_inputs_in_last_subtask",
        "all_inputs_in_last_test",
    )

    for section in config.sections():
        if re.match(r"test\d{2}", section):
            maybe_delete_key(config, section, "file_name")
        if re.match(r"solution_(.+)", section):
            maybe_rename_key(config, section, section, "subtasks", "tests")
            maybe_rename_key(config, section, section, "source", "run")
            maybe_rename_key(config, section, section, "points_above", "points_min")
            maybe_rename_key(config, section, section, "points_below", "points_max")

    maybe_rename_key(config, "task", "run_solution", "solutions_subdir", "subdir")

    maybe_move_key(config, "name", "all_tests", "tests")
    maybe_move_key(config, "in_globs", "all_tests", "tests")
    maybe_move_key(config, "predecessors", "all_tests", "tests")


OUTDATED_VERSIONS = {"v1": ("v2", update_to_v2), "v2": ("v3", update_to_v3)}
NEWEST_VERSION = "v3"


def update_config(config: ConfigParser, task_path: str, infos: bool = True) -> None:
    def inform(msg: str):
        if infos:
            eprint(ColorSettings.colored(msg, "yellow"))

    version = config.get("task", "version", fallback="v1")
    if version == NEWEST_VERSION:
        return

    if version not in OUTDATED_VERSIONS:
        raise TaskConfigError(f"Unknown version of config: {version}")

    inform(f"Updating config")
    while version in OUTDATED_VERSIONS:
        version, updater = OUTDATED_VERSIONS[version]
        updater(config, task_path)

    if version != NEWEST_VERSION:
        raise RuntimeError("Config updating failed.")

    config["task"]["version"] = NEWEST_VERSION



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import abstractmethod

from pisek.env.env import Env
from pisek.utils.paths import IInputPath
from pisek.config.task_config import ProgramRole, RunSection
from pisek.task_jobs.run_result import RunResult
from pisek.task_jobs.program import ProgramsJob


class ValidatorJob(ProgramsJob):
    """Runs validator on single input. (Abstract class)"""

    def __init__(
        self,
        env: Env,
        validator: RunSection,
        input_: IInputPath,
        test: int,
        **kwargs,
    ):
        super().__init__(env=env, name=f"Validate {input_:n} on test {test}", **kwargs)
        self.validator = validator
        self.test = test
        self.input = input_
        self.log_file = input_.to_log(f"{validator.name}{test}")

    @property
    @abstractmethod
    def _expected_returncode(self) -> int:
        pass

    @abstractmethod
    def _validation_args(self) -> list[str]:
        pass

    def _validate(self) -> RunResult:
        return self._run_program(
            ProgramRole.validator,
            self.validator,
            args=self._validation_args(),
            stdin=self.input,
            stderr=self.log_file,
        )

    def _run(self) -> None:
        result = self._validate()
        if result.returncode != self._expected_returncode:
            raise self._create_program_failure(
                f"Validator failed on {self.input:p} (test {self.test}):",
                result,
                stderr_force_content=True,
            )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>

from enum import auto, StrEnum
from pydantic_core import PydanticCustomError
from pydantic import BeforeValidator
from typing import Annotated, Literal


class TaskType(StrEnum):
    batch = auto()
    interactive = auto()


class OutCheck(StrEnum):
    diff = auto()
    tokens = auto()
    shuffle = auto()
    judge = auto()


class GenType(StrEnum):
    opendata_v1 = "opendata-v1"
    cms_old = "cms-old"
    pisek_v1 = "pisek-v1"


class ValidatorType(StrEnum):
    simple_0 = "simple-0"
    simple_42 = "simple-42"


class JudgeType(StrEnum):
    cms_batch = "cms-batch"
    cms_communication = "cms-communication"
    opendata_v1 = "opendata-v1"
    opendata_v2 = "opendata-v2"


class ShuffleMode(StrEnum):
    lines = auto()
    words = auto()
    lines_words = auto()
    tokens = auto()


class DataFormat(StrEnum):
    text = auto()
    strict_text = "strict-text"
    binary = auto()


def validate_test_points(points: str):
    if points == "unscored":
        return "unscored"
    try:
        p = int(points)
        assert p >= 0
        return p
    except (AssertionError, ValueError):
        raise PydanticCustomError(
            "test_points_parsing",
            "Input should be non-negative integer or 'unscored'",
        )


TestPoints = Annotated[int | Literal["unscored"], BeforeValidator(validate_test_points)]


class ProgramRole(StrEnum):
    gen = auto()
    validator = auto()
    primary_solution = auto()
    secondary_solution = auto()
    judge = auto()

    def is_solution(self) -> bool:
        return self in (ProgramRole.primary_solution, ProgramRole.secondary_solution)

    @property
    def build_name(self) -> str:
        if self.is_solution():
            return "solution"
        else:
            return self.name


class BuildStrategyName(StrEnum):
    python = auto()
    shell = auto()
    c = auto()
    cpp = auto()
    pascal = auto()
    java = auto()
    make = auto()
    cargo = auto()
    auto = auto()


class CMSFeedbackLevel(StrEnum):
    full = auto()
    restricted = auto()
    oi_restricted = auto()


class CMSScoreMode(StrEnum):
    max = auto()
    max_subtask = auto()
    max_tokened_last = auto()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from enum import StrEnum, auto
from typing import Any

from pisek.env.env import Env
from pisek.task_jobs.task_job import TaskJob
from pisek.utils.paths import (
    TESTS_DIR,
    INPUTS_LIST,
    InputPath,
    IInputPath,
    IOutputPath,
    OutputPath,
    TaskPath,
)


class TestcaseGenerationMode(StrEnum):
    static = auto()
    mixed = auto()
    generated = auto()


@dataclass(frozen=True, order=True)
class TestcaseInfo:
    generation_mode: TestcaseGenerationMode
    name: str
    repeat: int = 1
    seeded: bool = True

    @staticmethod
    def generated(name: str, repeat: int = 1, seeded: bool = True) -> "TestcaseInfo":
        return TestcaseInfo(TestcaseGenerationMode.generated, name, repeat, seeded)

    @staticmethod
    def mixed(name: str) -> "TestcaseInfo":
        return TestcaseInfo(TestcaseGenerationMode.mixed, name, 1, False)

    @staticmethod
    def static(name: str) -> "TestcaseInfo":
        return TestcaseInfo(TestcaseGenerationMode.static, name, 1, False)

    def input_path(
        self, seed: int | None = None, solution: str | None = None
    ) -> IInputPath:
        filename = self.name
        if self.seeded:
            assert seed is not None
            filename += f"_{seed:016x}"
        filename += ".in"

        return InputPath.new(filename, solution=solution)

    def reference_output(
        self, env: Env, seed: int | None = None, solution: str | None = None
    ) -> IOutputPath:
        is_static = self.generation_mode == TestcaseGenerationMode.static

        input_path = self.input_path(seed, solution=env.config.primary_solution)
        path: IOutputPath
        if is_static:
            path = OutputPath.static(input_path.replace_suffix(".out").name)
        else:
            path = input_path.to_output()

        if solution is not None:
            path = OutputPath(TESTS_DIR, solution, path.name).to_reference_output()

        return path

    def to_str(self) -> str:
        return f"{self.generation_mode} {self.name} repeat={self.repeat} seeded={str(self.seeded).lower()}"

    @staticmethod
    def from_str(line: str) -> "TestcaseInfo":
        line = line.rstrip("\n")
        if not line:
            raise ValueError("Line empty")

        args = line.split(" ")

        generation_mode = TestcaseGenerationMode(args[0])
        input_name = args[1]
        info_args: dict[str, Any] = {}

        for arg in args[2:]:
            if not arg.strip():
                raise ValueError("Argument empty. (Check whitespace.)")

            if "=" not in arg:
                raise ValueError("Missing '='")
            parts = arg.split("=")
            if len(parts) != 2:
                raise ValueError("Too many '='")
            arg_name, arg_value = parts
            if arg_name in info_args:
                raise ValueError(f"Repeated key '{arg_name}'")
            elif arg_name == "repeat":
                try:
                    repeat_times = int(arg_value)
                    assert repeat_times > 0
                except (ValueError, AssertionError):
                    raise ValueError("'repeat' should be a positive number")

                info_args[arg_name] = repeat_times
            elif arg_name == "seeded":
                if arg_value not in ("true", "false"):
                    raise ValueError("'seeded' should be 'true' or 'false'")
                info_args[arg_name] = arg_value == "true"
            else:
                raise ValueError(f"Unknown argument: '{arg_name}'")

        if not info_args.get("seeded", True) and info_args.get("repeat", 1) > 1:
            raise ValueError("For an unseeded input 'repeat' must be '1'")

        return TestcaseInfo(generation_mode, input_name, **info_args)


class ExportInputsList(TaskJob):
    def __init__(self, env: Env, testcases: list[TestcaseInfo]) -> None:
        self._testcases = testcases
        super().__init__(env=env, name=f"Export inputs list")

    def _run(self) -> None:
        assert list(sorted(self._testcases)) == self._testcases
        with self._open_file(TaskPath(TESTS_DIR, INPUTS_LIST), "w") as f:
            for testcase in self._testcases:
                f.write(testcase.to_str() + "\n")



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from pisek.env.env import Env
from pisek.config.config_types import ProgramRole
from pisek.config.task_config import RunSection
from pisek.utils.paths import IInputPath
from pisek.task_jobs.program import ProgramsJob, RunResultKind
from pisek.task_jobs.data.testcase_info import TestcaseInfo

from .base_classes import GeneratorListInputs, GenerateInput, GeneratorTestDeterminism


class OpendataV1ListInputs(GeneratorListInputs):
    """Lists all inputs for opendata-v1 generator - one for each test."""

    def __init__(self, env: Env, generator: RunSection, **kwargs) -> None:
        super().__init__(env=env, generator=generator, **kwargs)

    def _run(self) -> list[TestcaseInfo]:
        # Although for test in self._env.config.test_sections
        # would be shorter, this doesn't actually access env variables
        # and that leads to caching bug.
        return [
            TestcaseInfo.generated(f"{test:02}")
            for test in range(self._env.config.tests_count)
            if test != 0
        ]


class OpendataV1GeneratorJob(ProgramsJob):
    """Abstract class for jobs with OnlineGenerator."""

    generator: RunSection
    seed: Optional[int]
    testcase_info: TestcaseInfo
    input_path: IInputPath

    def __init__(self, env: Env, *, name: str = "", **kwargs) -> None:
        super().__init__(env=env, name=name, **kwargs)

    def _gen(self) -> None:
        assert self.seed is not None
        if self.seed < 0:
            raise ValueError(f"seed {self.seed} is negative")

        test = int(self.testcase_info.name)

        result = self._run_program(
            ProgramRole.gen,
            self.generator,
            args=[str(test), f"{self.seed:016x}"],
            stdout=self.input_path.to_raw(self._env.config.tests.in_format),
            stderr=self.input_path.to_log(self.generator.name),
        )
        if result.kind != RunResultKind.OK:
            raise self._create_program_failure(
                f"{self.generator.name} failed on test {test}, seed {self.seed:016x}:",
                result,
                stderr_force_content=True,
            )


class OpendataV1Generate(OpendataV1GeneratorJob, GenerateInput):
    """Generates input with given name."""

    pass


class OpendataV1TestDeterminism(OpendataV1GeneratorJob, GeneratorTestDeterminism):
    """Tests determinism of generating a given input."""

    pass



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from decimal import Decimal
import subprocess
from typing import Optional

from pisek.env.env import Env
from pisek.utils.paths import IInputPath, IOutputPath
from pisek.jobs.jobs import PipelineItemFailure
from pisek.utils.text import tab
from pisek.task_jobs.run_result import RunResult, RunResultKind
from pisek.task_jobs.solution.solution_result import (
    Verdict,
    SolutionResult,
    RelativeSolutionResult,
)
from pisek.task_jobs.checker.checker_base import RunBatchChecker


class RunDiffChecker(RunBatchChecker):
    """Checks solution output and correct output using diff."""

    def __init__(
        self,
        env: Env,
        test: int,
        input_: IInputPath,
        output: IOutputPath,
        correct_output: IOutputPath,
        expected_verdict: Optional[Verdict],
    ) -> None:
        super().__init__(
            env=env,
            checker_name="diff",
            test=test,
            input_=input_,
            output=output,
            correct_output=correct_output,
            expected_verdict=expected_verdict,
        )

    def _check(self) -> SolutionResult:
        self._access_file(self.output)
        self._access_file(self.correct_output)
        diff = self._run_subprocess(
            ["diff", "-Bbq", self.output.path, self.correct_output.path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        # XXX: Okay, it didn't finish in no time, but this is not meant to be used
        rr = RunResult(
            RunResultKind.OK,
            diff.returncode,
            0,
            0,
            status=("Files are the same" if diff.returncode == 0 else "Files differ")
            + f": {self.output.col(self._env)} {self.correct_output.col(self._env)}",
        )
        if diff.returncode == 0:
            return RelativeSolutionResult(
                verdict=Verdict.ok,
                message=None,
                solution_rr=self._solution_run_res,
                checker_rr=rr,
                relative_points=Decimal(1),
            )
        elif diff.returncode == 1:
            return RelativeSolutionResult(
                verdict=Verdict.wrong_answer,
                message=None,
                solution_rr=self._solution_run_res,
                checker_rr=rr,
                relative_points=Decimal(0),
            )
        else:
            assert diff.stderr is not None
            raise PipelineItemFailure(
                f"Diff failed:\n{tab(diff.stderr.read().decode('utf-8'))}"
            )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from pisek.env.env import Env
from pisek.utils.paths import IInputPath, IOutputPath
from pisek.config.config_types import OutCheck, JudgeType
from pisek.task_jobs.solution.solution_result import Verdict

from pisek.task_jobs.checker.checker_base import RunBatchChecker
from pisek.task_jobs.checker.diff_checker import RunDiffChecker
from pisek.task_jobs.checker.judgelib_checker import RunTokenChecker, RunShuffleChecker
from pisek.task_jobs.checker.cms_judge import RunCMSBatchJudge
from pisek.task_jobs.checker.opendata_judge import (
    RunOpendataV1Judge,
    RunOpendataV2Judge,
)


def checker_job(
    input_: IInputPath,
    output: IOutputPath,
    correct_output: IOutputPath,
    test: int,
    seed: Optional[int],
    expected_verdict: Optional[Verdict],
    env: Env,
) -> RunBatchChecker:
    """Returns JudgeJob according to contest type."""
    if env.config.tests.out_check == OutCheck.diff:
        return RunDiffChecker(
            env, test, input_, output, correct_output, expected_verdict
        )

    if env.config.tests.out_check == OutCheck.tokens:
        return RunTokenChecker(
            env, test, input_, output, correct_output, expected_verdict
        )
    elif env.config.tests.out_check == OutCheck.shuffle:
        return RunShuffleChecker(
            env, test, input_, output, correct_output, expected_verdict
        )

    if env.config.tests.out_judge is None:
        raise ValueError(f"Unset judge for out_check={env.config.tests.out_check.name}")

    if env.config.tests.judge_type == JudgeType.cms_batch:
        return RunCMSBatchJudge(
            env,
            env.config.tests.out_judge,
            test,
            input_,
            output,
            correct_output,
            expected_verdict,
        )
    elif env.config.tests.judge_type == JudgeType.opendata_v1:
        return RunOpendataV1Judge(
            env,
            env.config.tests.out_judge,
            test,
            input_,
            output,
            correct_output,
            seed,
            expected_verdict,
        )
    elif env.config.tests.judge_type == JudgeType.opendata_v2:
        return RunOpendataV2Judge(
            env,
            env.config.tests.out_judge,
            test,
            input_,
            output,
            correct_output,
            seed,
            expected_verdict,
        )
    else:
        raise ValueError(f"Unknown judge type: {env.config.tests.judge_type}")



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from pisek.utils.text import tab, pad
from pisek.utils.terminal import MSG_LEN, TARGET_LINE_WIDTH
from pisek.jobs.jobs import State, PipelineItem, JobManager

TARGET_BAR_WIDTH = TARGET_LINE_WIDTH - 11


class StatusJobManager(JobManager):
    """JobManager that implements useful methods for terminal interaction."""

    def _bar(self, msg: str, part: int, full: int, color: str = "cyan") -> str:
        """Return progress bar with given parameters."""
        msg = pad(msg, MSG_LEN - 1) + " "
        progress_msg = f"  ({part}/{full})"

        bar_len = min(TARGET_BAR_WIDTH, TARGET_LINE_WIDTH - len(progress_msg)) - len(
            msg
        )
        filled = bar_len * part // full

        if filled in (0, bar_len):
            bar = "━" * filled
        else:
            bar = (filled - 1) * "━" + "╸"

        return f"{msg}{self._colored(bar, color)}{self._colored((bar_len-filled)*'━', 'white')}{progress_msg}"

    def _job_bar(self, msg: str) -> str:
        """Returns progress bar according to status of this manager's jobs."""
        color = "cyan"
        if self.state == State.cancelled:
            return f"{pad(msg, MSG_LEN-1)} {self._colored('cancelled', 'yellow')}"
        elif self.state == State.succeeded:
            color = "green"
        elif self.state == State.failed or State.failed in self._job_states():
            color = "red"

        return self._bar(
            msg,
            len(self.jobs_with_state(State.succeeded))
            + (self.state == State.succeeded),
            len(self.jobs) + 1,
            color=color,
        )

    def get_status(self) -> str:
        return self._job_bar(self.name)

    @staticmethod
    def _fail_message(pitem: PipelineItem) -> str:
        """Get fail message of given job."""
        return f'"{pitem.name}" failed:\n{tab(pitem.fail_msg)}\n'



#!/usr/bin/env python3
import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Loads an input from stdin and validates it."
    )
    parser.add_argument("test", type=int, help="test number (0-2)")
    args = parser.parse_args()

    # Here the bounds for test 3 are stricter than what the generator really creates.
    BOUNDS = [(-1e18, 1e18), (-1e9, 1e9), (-1e18, 1e18)]
    assert 0 <= args.test < len(BOUNDS)
    bounds = BOUNDS[args.test]

    t = int(input())
    for _ in range(t):
        a, b = map(int, input().split(" "))
        assert bounds[0] <= a <= bounds[1]
        assert bounds[0] <= b <= bounds[1]

    end = True
    try:
        input()
    except EOFError:
        end = True
    assert end
    exit(42)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import abstractmethod
from decimal import Decimal
from typing import Optional
from functools import cache

from pisek.utils.text import tab
from pisek.utils.paths import IInputPath, IOutputPath, LogPath
from pisek.env.env import Env
from pisek.config.config_types import DataFormat
from pisek.task_jobs.tools import SanitizationResultKind
from pisek.jobs.jobs import State, PipelineItemFailure
from pisek.task_jobs.run_result import RunResult, RunResultKind
from pisek.task_jobs.program import ProgramsJob
from pisek.task_jobs.solution.solution_result import (
    Verdict,
    SolutionResult,
    RelativeSolutionResult,
    AbsoluteSolutionResult,
)


class RunChecker(ProgramsJob):
    """Runs checker on single input. (Abstract class)"""

    def __init__(
        self,
        env: Env,
        name: str,
        test: int,
        checker_name: str,
        input_: IInputPath,
        checker_log_file: LogPath,
        expected_verdict: Optional[Verdict],
        **kwargs,
    ) -> None:
        super().__init__(env=env, name=name, **kwargs)
        self.test = test
        self.input = input_
        self.checker_name = checker_name
        self.checker_log_file = checker_log_file
        self.expected_verdict = expected_verdict

        self.result: Optional[SolutionResult]

    @cache
    def _load_solution_run_res(self) -> None:
        """
        Loads solution's RunResult into self.solution_res
        """
        self._solution_run_res = self._get_solution_run_res()

    @abstractmethod
    def _get_solution_run_res(self) -> RunResult:
        """
        Gets solution's RunResult.
        Call this only through _load_solution_run_res as this can run the solution.
        """
        pass

    @abstractmethod
    def _check(self) -> SolutionResult:
        """Here actually do the output checking."""
        pass

    @abstractmethod
    def _checking_message(self) -> str:
        pass

    def _checking_message_capitalized(self) -> str:
        msg = self._checking_message()
        return msg[0].upper() + msg[1:]

    def _run(self) -> SolutionResult:
        self._load_solution_run_res()
        result = self._get_solution_result()

        if (
            self.expected_verdict is not None
            and result.verdict != self.expected_verdict
        ):
            raise PipelineItemFailure(
                f"{self._checking_message_capitalized()} should have got verdict '{self.expected_verdict}' but got '{result.verdict}'."
            )

        return result

    def _get_solution_result(self) -> SolutionResult:
        if self._solution_run_res.kind == RunResultKind.OK:
            out_form = self._env.config.tests.out_format
            san_res = self.prerequisites_results.get("sanitize")
            if san_res is None:
                pass
            elif san_res.kind == SanitizationResultKind.invalid:
                return RelativeSolutionResult(
                    verdict=Verdict.normalization_fail,
                    message=san_res.msg,
                    solution_rr=self._solution_run_res,
                    checker_rr=None,
                    relative_points=Decimal(0),
                )
            elif (
                out_form == DataFormat.strict_text
                and san_res.kind == SanitizationResultKind.changed
            ):
                return RelativeSolutionResult(
                    verdict=Verdict.normalization_fail,
                    message=f"Output not normalized. (Check encoding, missing newline at the end or '\\r'.)",
                    solution_rr=self._solution_run_res,
                    checker_rr=None,
                    relative_points=Decimal(0),
                )

            return self._check()
        elif self._solution_run_res.kind == RunResultKind.RUNTIME_ERROR:
            return RelativeSolutionResult(
                verdict=Verdict.error,
                message=None,
                solution_rr=self._solution_run_res,
                checker_rr=None,
                relative_points=Decimal(0),
            )
        elif self._solution_run_res.kind == RunResultKind.TIMEOUT:
            return RelativeSolutionResult(
                verdict=Verdict.timeout,
                message=None,
                solution_rr=self._solution_run_res,
                checker_rr=None,
                relative_points=Decimal(0),
            )

        raise ValueError(
            f"Invalid solution run result kind: '{self._solution_run_res.kind}'"
        )

    def message(self) -> str:
        """Message about how checking ended."""
        if self.result is None:
            raise RuntimeError(f"Job {self.name} has not finished yet.")

        sol_rr = self.result.solution_rr
        checker_rr = self.result.checker_rr

        text = f"input: {self._quote_file_with_name(self.input)}"
        if isinstance(self, RunBatchChecker):
            text += f"correct output: {self._quote_file_with_name(self.correct_output)}"
        text += f"result: {self.result.verdict.name}\n"

        text += "solution:\n"
        text += tab(
            self._format_run_result(
                sol_rr,
                status=sol_rr.kind != RunResultKind.OK,
                stderr_force_content=sol_rr.kind == RunResultKind.RUNTIME_ERROR,
                time=True,
            )
        )
        text += "\n"
        if checker_rr is not None:
            text += (
                f"{self.checker_name}:\n"
                + tab(
                    self._format_run_result(
                        checker_rr,
                        status=checker_rr.stderr_file is None,
                        stderr_force_content=True,
                    )
                )
                + "\n"
            )

        return text

    def verdict_text(self) -> str:
        if self.result is not None:
            if self.result.message is not None:
                return self.result.message
            return self.result.verdict.name
        else:
            return self.state.name

    def verdict_mark(self) -> str:
        if self.state == State.cancelled:
            return "-"
        elif self.result is None:
            return " "
        else:
            return self.result.verdict.mark


class RunBatchChecker(RunChecker):
    """Runs batch checker on single input. (Abstract class)"""

    def __init__(
        self,
        env: Env,
        checker_name: str,
        test: int,
        input_: IInputPath,
        output: IOutputPath,
        correct_output: IOutputPath,
        expected_verdict: Optional[Verdict],
        **kwargs,
    ) -> None:
        super().__init__(
            env=env,
            name=f"Check {output:p}",
            checker_name=checker_name,
            test=test,
            input_=input_,
            checker_log_file=output.to_checker_log(checker_name),
            expected_verdict=expected_verdict,
            **kwargs,
        )
        self.output = output
        self.correct_output_name = correct_output
        self.correct_output = correct_output

    def _get_solution_run_res(self) -> RunResult:
        if "run_solution" in self.prerequisites_results:
            return self.prerequisites_results["run_solution"]
        else:
            # There is no solution (checking static tests against themselves)
            # XXX: It didn't technically finish in 0 time.
            return RunResult(RunResultKind.OK, 0, 0.0, 0.0)

    def _checking_message(self) -> str:
        return (
            f"output {self.output.col(self._env)} for input {self.input.col(self._env)}"
        )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import filecmp

from pisek.utils.paths import TaskPath
from pisek.env.env import TestingTarget
from pisek.config.config_types import JudgeType, TaskType
from pisek.task_jobs.solution.solution_result import Verdict
from pisek.jobs.jobs import Job
from pisek.task_jobs.task_manager import (
    TaskJobManager,
    FUZZ_MAN_CODE,
    SOLUTION_MAN_CODE,
)


class CompletenessCheck(TaskJobManager):
    """Checks task as a whole."""

    def __init__(self):
        super().__init__("Completeness check")

    def _get_jobs(self) -> list[Job]:
        return []

    def _get_checker_outs(self) -> set[TaskPath]:
        checker_outs: set[TaskPath] = set()
        if FUZZ_MAN_CODE in self.prerequisites_results:
            checker_outs |= self.prerequisites_results[FUZZ_MAN_CODE]["checker_outs"]
        for solution in self._env.solutions:
            checker_outs |= self.prerequisites_results[
                f"{SOLUTION_MAN_CODE}{solution}"
            ]["checker_outs"]
        return checker_outs

    def _check_solution_succeeds_only_on(self, sol_name: str, tests: list[int]) -> bool:
        tests_res = self.prerequisites_results[f"{SOLUTION_MAN_CODE}{sol_name}"][
            "tests"
        ]
        for num in self._env.config.test_sections:
            if num == 0:
                continue  # Skip samples
            if (tests_res[num] == Verdict.ok) != (num in tests):
                return False
        return True

    def _check_dedicated_solutions(self) -> None:
        """Checks that each test has it's own dedicated solution."""
        if self._env.config.checks.solution_for_each_test:
            for num, test in self._env.config.test_sections.items():
                if num == 0:
                    continue  # Samples

                ok = False
                for solution in self._env.solutions:
                    if self._check_solution_succeeds_only_on(
                        solution, [num] + test.all_predecessors
                    ):
                        ok = True
                        break

                if not ok:
                    self._warn(f"{test.name} has no dedicated solution")

    def _check_different_outputs(self) -> None:
        """Checks that for each test primary solution's outputs aren't the same."""
        if self._env.config.task.task_type == TaskType.interactive:
            return

        res = self.prerequisites_results[
            f"{SOLUTION_MAN_CODE}{self._env.config.primary_solution}"
        ]["results"]
        for test in self._env.config.test_sections.values():
            if not test.checks_different_outputs:
                continue

            all_same = True
            outs = [inp.to_output() for inp in res if test.in_test(inp.name)]
            if len(outs) <= 1:
                continue

            for out in outs:
                all_same &= filecmp.cmp(outs[0].path, out.path, shallow=False)

            if all_same:
                self._warn(f"All outputs of {test.name} are the same")

    def _check_cms_judge(self) -> None:
        """Checks that cms judge's stdout & stderr contains only one line."""
        if self._env.config.tests.judge_type in (
            JudgeType.cms_batch,
            JudgeType.cms_communication,
        ):
            for judge_out in self._get_checker_outs():
                with open(judge_out.path) as f:
                    lines = f.read().rstrip().split("\n")
                if len(lines) > 1 or lines[0] == "":  # Python splitting is weird
                    self._warn(f"{judge_out:p} should contain exactly one line")
                    if self._env.verbosity <= 0:
                        return

    def _evaluate(self) -> None:
        assert self._env.target == TestingTarget.all

        self._check_cms_judge()
        self._check_different_outputs()
        self._check_dedicated_solutions()



from dataclasses import dataclass
from decimal import Decimal

from pisek.task_jobs.data.testcase_info import TestcaseGenerationMode, TestcaseInfo
from pisek.task_jobs.solution.solution_result import Verdict


@dataclass
class OpendataTestcaseInfo:
    mode: TestcaseGenerationMode
    name: str
    repeat: int
    seeded: bool

    @staticmethod
    def from_testcase_info(testcase_info: TestcaseInfo) -> "OpendataTestcaseInfo":
        return OpendataTestcaseInfo(
            testcase_info.generation_mode,
            testcase_info.name,
            testcase_info.repeat,
            testcase_info.seeded,
        )


@dataclass
class OpendataVerdict:
    verdict: Verdict
    message: str | None
    points: Decimal
    log: str | None
    note: str | None
